Package: fenr
Title: Fast functional enrichment for interactive applications
Version: 1.10.1
Authors@R: person(
    given = "Marek",
    family = "Gierlinski",
    email = "M.Gierlinski@dundee.ac.uk",
    role = c("aut", "cre"),
    comment = c(ORCID = "0000-0001-9149-3514")
    ) 
Description: Perform fast functional enrichment on feature lists (like genes
    or proteins) using the hypergeometric distribution. Tailored for speed,
    this package is ideal for interactive platforms such as Shiny. It supports
    the retrieval of functional data from sources like GO, KEGG, Reactome,
    Bioplanet and WikiPathways. By downloading and preparing data first, it
    allows for rapid successive tests on various feature selections without
    the need for repetitive, time-consuming preparatory steps typical of other
    packages. 
URL: https://github.com/bartongroup/fenr
BugReports: https://github.com/bartongroup/fenr/issues
Depends: R (>= 4.1.0)
License: MIT + file LICENSE
Encoding: UTF-8
biocViews: FunctionalPrediction, DifferentialExpression,
        GeneSetEnrichment, GO, KEGG, Reactome, Proteomics
Imports: tools, methods, assertthat, rlang, dplyr, tidyr, tidyselect,
        tibble, purrr, readr, stringr, httr2, rvest, progress,
        BiocFileCache, shiny, ggplot2
Suggests: BiocStyle, testthat, knitr, rmarkdown, topGO
VignetteBuilder: knitr
LazyData: false
Config/roxygen2/version: 8.0.0
git_url: https://git.bioconductor.org/packages/fenr
git_branch: RELEASE_3_23
git_last_commit: 6578bfb
git_last_commit_date: 2026-06-30
Repository: Bioconductor 3.23
Date/Publication: 2026-07-02
NeedsCompilation: no
Packaged: 2026-07-02 21:48:17 UTC; biocbuild
Author: Marek Gierlinski [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9149-3514>)
Maintainer: Marek Gierlinski <M.Gierlinski@dundee.ac.uk>

# Generated by roxygen2: do not edit by hand

export(enrichment_interactive)
export(fetch_bp)
export(fetch_go)
export(fetch_go_species)
export(fetch_kegg)
export(fetch_kegg_species)
export(fetch_reactome)
export(fetch_reactome_species)
export(fetch_terms_for_example)
export(fetch_wiki)
export(fetch_wiki_species)
export(functional_enrichment)
export(get_feature_terms)
export(get_term_features)
export(prepare_for_enrichment)
export(remove_cache)
importFrom(assertthat,assert_that)
importFrom(assertthat,is.count)
importFrom(assertthat,is.string)
importFrom(methods,is)

# Fast functional enrichment

Maintainer: Marek Gierlinski (<M.Gierlinski@dundee.ac.uk>)

This R package provides a fast and efficient method for functional enrichment analysis, optimized for speed and designed for use in interactive applications, such as *Shiny* apps.


## Quick example

To run a *Shiny* app demonstration of `fenr` directly from GitHub, enter the following command in your R console:

```
shiny::runGitHub("bartongroup/fenr-shiny-example")
```

## Installation

`fenr` can be installed using

```
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("fenr")
```

## Usage

The initial step involves downloading functional term data. `fenr` supports data downloads from *Gene Ontology*, *Reactome*, *KEGG*, and *WikiPathways*. Custom ontologies can also be used, provided they are converted into an appropriate format (refer to the `prepare_for_enrichment` function for more information). The command below downloads functional terms and gene mapping from Gene Ontology (GO) for yeast:

```
go <- fetch_go(species = "sgd")
```

This command returns a list with two tibbles containing term information (`term_id` and `term_name`) and gene-term mapping (`term_id` and `gene_symbol`). We convert this data into an object suitable for fast functional enrichment. `exmpl_all` is an example of gene background provided by the package, which contains a vector with gene symbols related to all detections in an experiment.

```
data(exmpl_all, exmpl_sel)
go_terms <- prepare_for_enrichment(go$terms, go$mapping, exmpl_all, feature_name = "gene_symbol")
```

The `go_terms` object is a data structure containing all mappings in a quickly accessible form. From this point on, you can use go_terms to perform multiple functional enrichments on various gene selections. For example, if `exmpl_all` is a vector with all background gene symbols and `exmpl_sel` is a vector with genes of interest (both provided by the package), you can perform functional enrichment analysis using:

```
enr <- functional_enrichment(exmpl_all, exmpl_sel, go_terms)
```

The result is a tibble:

```
# A tibble: 51 × 10
  term_id    term_name                                N_with n_with_sel n_expect enrichment odds_ratio ids    p_value p_adjust
   <chr>      <chr>                                     <int>      <int>    <dbl>      <dbl>      <dbl> <chr>    <dbl>    <dbl>
 1 GO:1905356 regulation of snRNA pseudouridine synth…      2          2     0.01      333        Inf   TOR2… 8.61e- 6 5.49e- 5
 2 GO:0031929 TOR signaling                                19         18     0.06      315      41800   TOR2… 0        0       
 3 GO:0031931 TORC1 complex                                11         10     0.03      302       6330   TOR2… 0        0       
 4 GO:0001558 regulation of cell growth                     9          5     0.03      185        544   KOG1… 1.84e-11 2.34e-10
 5 GO:0031932 TORC2 complex                                15          7     0.05      155        435   TOR2… 4.66e-15 7.93e-14
 6 GO:0016242 negative regulation of macroautophagy         9          3     0.03      111        193   KSP1… 1.95e- 6 1.42e- 5
 7 GO:0043666 regulation of phosphoprotein phosphatas…     13          4     0.04      102        182   SAP1… 4.24e- 8 4.33e- 7
 8 GO:0030950 establishment or maintenance of actin c…     15          4     0.05       88.7      149   TOR2… 8.07e- 8 6.86e- 7
 9 GO:0031930 mitochondria-nucleus signaling pathway        9          2     0.03       73.9      105   TOR1… 3.06e- 4 1.3 e- 3
10 GO:0010507 negative regulation of autophagy             12          2     0.04       55.4       73.2 TOR2… 5.58e- 4 2.03e- 3
# ℹ 41 more rows
```

The columns are as follows

 - `N_with`: The number of features (genes) associated with this term in the background of all genes.
 - `n_with_sel`: The number of features associated with this term in the selection.
 - `n_expect`: The expected number of features associated with this term under the null hypothesis (terms are randomly distributed).
 - `enrichment`: The ratio of observed to expected.
 - `odds_ratio`: The effect size, represented by the odds ratio from the contingency table.
 - `ids`: The identifiers of features with the term in the selection.
 - `p_value`: The raw p-value from the hypergeometric distribution.
 - `p_adjust`: The p-value adjusted for multiple tests using the Benjamini-Hochberg approach.

# Interactive Example

A small Shiny app is included in the package to demonstrate the usage of `fenr` in an interactive environment. All time-consuming data loading and preparation tasks are performed before the app is launched.

```
data(yeast_de)
term_data <- fetch_terms_for_example(yeast_de)
```
 
`yeast_de` is the result of differential expression (using `edgeR`) on a subset of 6+6 replicates from [Gierlinski et al. (2015)](https://academic.oup.com/bioinformatics/article/31/22/3625/240923).

The function `fetch_terms_for_example` uses `fetch_*` functions from `fenr` to download and process data from *GO*, *Reactome* and *KEGG*.  You can view the step-by-step process by examining the function code on [GitHub](https://github.com/bartongroup/fenr/blob/main/R/iteractive_example.R). The object `term_data` is a named list of `fenr_terms` objects, one for each ontology.

After completing the slow tasks, you can start the Shiny app by running:

```
enrichment_interactive(yeast_de, term_data)
```

To quickly see how `fenr` works an example can be loaded directly from GitHub:

```
shiny::runGitHub("bartongroup/fenr-shiny-example")
```

#' WikiPathways server base URL
#'
#' @return A string with URL. This can be changed by options(WIKI_BASE_URL = "different/url").
#' @noRd
get_wiki_url <- function() {
  getOption("WIKI_BASE_URL", "https://webservice.wikipathways.org")
}

#' List of available WikiPathways species
#'
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A character vector with species names used by WikiPathways.
#' @export
#' @examples
#' spec <- fetch_wiki_species(on_error = "warn")
fetch_wiki_species <- function(on_error = c("stop", "warn", "ignore")) {
  on_error <- match.arg(on_error)

  resp <- http_request(get_wiki_url(), "listOrganisms", parameters = list(format = "json"))
  if(resp$is_error)
    return(catch_error("WikiPathways", resp, on_error))

  js <- httr2::resp_body_json(resp$response)
  tibble::tibble(designation = unlist(js))
}

#' Download pathway data from WikiPathways
#'
#' @param species Species name recognised by WikiPathways (see \code{fetch_wiki_species()})
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A tibble with columns \code{gene_id} and \code{term_id}
#' @noRd
fetch_wiki_pathways <- function(species, on_error) {
  id <- name <- NULL

  resp <- http_request(get_wiki_url(), "listPathways",
                       parameters = list(format = "json", organism = species))
  if(resp$is_error)
    return(catch_error("WikiPathways", resp, on_error))

  js <- httr2::resp_body_json(resp$response)
  # Sometimes WikiPathways server responds OK, but returns "error" text
  if(is(js[[1]], "character") && js[[1]] == "error") {
    msg <- stringr::str_glue("Cannot retrieve pathways from WikiPathways for species {species}.")
    return(error_response(msg, on_error))
  }

  js$pathways |>
    purrr::map(tibble::as_tibble) |>
    purrr::list_rbind() |>
    dplyr::select(term_id = id, term_name = name)
}


#' Helper function for `parse_wiki_gpml`
#'
#' @param s Input string
#' @param key Key to extract
#'
#' @return Value of the key
#' @noRd
extract_key <- function(s, key) {
  stringr::str_extract(s, paste0("(?<=", key, '=\")(.+?)(?=")'))
}

#' Parse GPML string provided by WikiPathways
#'
#' @param gpml GMPL string returned by WikiPathways query.
#' @param keys Keys to be extracted.
#'
#' @return A tibble with selected keys and values
#' @noRd
parse_wiki_gpml <- function(gpml, keys = c("TextLabel", "Type", "Database", "ID")) {
  gpml |>
    stringr::str_replace_all("\\n", "") |>
    stringr::str_extract_all("<DataNode.+?</DataNode>") |>
    purrr::pluck(1) |>
    purrr::map_dfr(function(s) {
      purrr::map_chr(keys, ~extract_key(s, .x)) |>
        purrr::set_names(keys)
    })
}


#' Download WikiPathways gene-pathway mapping using API
#'
#' @param pathways A character vector with pathway names.
#' @param databases Names of databases to use
#' @param types Names of types to use
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A tibble with \code{term_id} and \code{gene_symbol}
#' @noRd
fetch_wiki_pathway_genes_api <- function(pathways, databases = NULL, types = NULL, on_error) {
  term_id <- TextLabel <-ID <- Type <- Database <- database <- type <- NULL

  raise_error <- FALSE
  pb <- progress::progress_bar$new(total = length(pathways))
  res <- purrr::map(pathways, function(pathway) {
    pb$tick()
    resp <- http_request(get_wiki_url(), "getPathway",
                         parameters = list(format = "json", pwId = pathway))
    if(resp$is_error) {
      catch_error("WikiPathways", resp, on_error)
      raise_error <<- TRUE
      return()
    }
    js <- httr2::resp_body_json(resp$response)
    parse_wiki_gpml(js$pathway$gpml) |>
      tibble::add_column(term_id = pathway)
  }) |>
    purrr::list_rbind()
  if(raise_error)
    return(NULL)

  res <- res |>
    dplyr::select(term_id, text_label = TextLabel, id = ID, type = Type, database = Database) |>
    dplyr::distinct()
  if(!is.null(databases)) {
    res <- res |>
      dplyr::filter(database %in% databases)
  }
  if(!is.null(types)) {
    res <- res |>
      dplyr::filter(type %in% types)
  }

  res
}


#' Get functional term data from WikiPathways
#'
#' Download term information (pathway ID and name) and gene-pathway mapping
#' (gene symbol and pathway ID) from WikiPathways.
#'
#' @details WikiPathways contain mapping between pathways and a variety of
#'   entities from various databases. Typically a gene symbol is returned in
#'   column \code{text_label} and some sort of ID in column \code{id}, but this
#'   depends on the species and databases used. For gene/protein enrichment,
#'   these should be filtered to contain gene symbols only. This can be done by
#'   selecting a desired databases and types. The default values for parameters
#'   \code{databases} and \code{types} attempt to select information from generic
#'   databases, but there are organism-specific databases not included in the
#'   selection. We suggest to run this function with \code{databases = NULL,
#'   types = NULL} to see what types and databases are available before making
#'   selection.
#'
#' @param species WikiPathways species designation, for example "Homo sapiens"
#'   for human. Full list of available species can be found using
#'   \code{fetch_wiki_species()}.
#' @param databases A character vector with database names to pre-filter mapping
#'   data. See details. Full result will be returned if NULL.
#' @param types A character vector with types of entities to pre-filter mapping
#'   data. See details. Full result will be returned if NULL.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A list with \code{terms} and \code{mapping} tibbles.
#' @export
#' @importFrom assertthat assert_that
#' @examples
#' wiki_data <- fetch_wiki("Bacillus subtilis", on_error = "warn")
fetch_wiki <- function(
    species,
    databases = c("Ensembl", "Entrez Gene", "HGNC", "HGNC Accession number", "Uniprot-TrEMBL"),
    types = c("GeneProduct", "Protein", "Rna", "RNA"),
    on_error = c("stop", "warn", "ignore")
  ) {
  on_error <- match.arg(on_error)
  assert_that(!missing(species), msg = "Argument 'species' is missing.")
  assert_species(species, "fetch_wiki_species", on_error)

  terms <- fetch_wiki_pathways(species, on_error)
  if(is.null(terms))
    return(NULL)
  mapping <- fetch_wiki_pathway_genes_api(terms$term_id, databases, types, on_error)
  list(
    terms = terms,
    mapping = mapping
  )
}

test_that("Incorrect species in fetch_wiki", {
  expect_error(fetch_wiki())
  expect_error(fetch_wiki(1243))
  expect_error(fetch_wiki("not a species"))
})


test_that("Incorrect species in fetch_wiki_pathways", {
  expect_error(fetch_wiki_pathways("Heffalump"))
})


test_that("Parsing GPLM", {
  expected <- tibble::tibble(
    TextLabel = c("STR3", "CYS4", "STR2"),
    Type = c("GeneProduct", "GeneProduct", "GeneProduct"),
    Database = c("SGD", "SGD", "SGD"),
    ID = c("S000003152", "S000003387", "S000003891")
  )

  gpml <- readLines("../test_data/wiki_test.gpml") |>
    paste(collapse = "\n")
  res <- parse_wiki_gpml(gpml)

  expect_equal(res, expected)
})

test_that("Expected behaviour from a non-responsive server", {
  species <- "Bacillus subtilis"
  pathway <- "WP1466"
  httr2::with_mocked_responses(
    mock = mocked_500,
    code = {
      test_unresponsive_server(fetch_wiki_species)
      test_unresponsive_server(fetch_wiki_pathways, species = species)
      test_unresponsive_server(fetch_wiki, species = species)
      test_unresponsive_server(fetch_wiki_pathway_genes_api, pathways = pathway)
      test_unresponsive_server(fetch_wiki, species = species)
    })
})

test_that("Expected return from fetch_wiki_species", {
  expected_selection <- c("Homo sapiens", "Mus musculus", "Rattus norvegicus",
                          "Saccharomyces cerevisiae", "Drosophila melanogaster", "Caenorhabditis elegans")
  spec <- fetch_wiki_species(on_error = "ignore")
  if(!is.null(spec)) {
    expect_is(spec, "tbl")
    expect_true(all(expected_selection %in% spec$designation))
  }
})


test_that("WikiPathways correct response", {
  species <- "Bacillus subtilis"
  databases <- c("Ensembl", "Entrez Gene", "HGNC", "HGNC Accession number", "Uniprot-TrEMBL")
  types <- c("GeneProduct", "Protein", "Rna", "RNA")

  expected_terms <- tibble::tribble(
    ~term_id, ~term_name,
    "WP1466", "Response regulator aspartate phosphatase interactions",
    "WP1527", "Stress response",
    "WP2360", "Folate biosynthesis"
  )

  expected_mapping <- tibble::tribble(
    ~term_id, ~text_label,
    "WP1466", "rapA",
    "WP1466", "rapE",
    "WP1527", "sinR",
    "WP1527", "sinI"
  )

  re <- fetch_wiki(species, databases = databases, types = types, on_error = "ignore")
  if(!is.null(re)) {
    test_fetched_structure(re)
    test_terms(re$terms, expected_terms)
    test_mapping(re$mapping, expected_mapping, "text_label")
  }
})

#' Retrieve Cache Location for `fenr`
#'
#' This function returns the cache directory location for the `fenr` package.
#'
#' @return A character string representing the path to the cache directory for
#'   `fenr`.
#' @noRd
cache_location <- function() {
  tools::R_user_dir("fenr", which = "cache")
}


#' URL Cache Path Location
#'
#' This function queries a BiocFileCache object for a given resource name
#' (`rname`). If the resource isn't found, it's added to the cache. If there's
#' one entry found, it checks for updates and retrieves the local path. If
#' multiple entries are found, it cleans up by removing them and then adds the
#' resource anew.
#'
#' @param bfc A BiocFileCache object.
#' @param rname A character string representing the resource name to be queried
#'   or added.
#' @param fpath A character string representing URL of the file.
#'
#' @return A character string representing the local path of the cached
#'   resource.
#' @noRd
cache_path <- function(bfc, rname, fpath) {
  hits <- BiocFileCache::bfcquery(bfc, rname, field = "rname")

  # Not found, add to cache, get local path
  if(nrow(hits) == 0) {
    lpath <- BiocFileCache::bfcadd(bfc, rname, fpath)
  }

  # Found one entry, check for updates, get local path
  else if(nrow(hits) == 1) {
    if(BiocFileCache::bfcneedsupdate(bfc, hits$rid))
      BiocFileCache::bfcupdate(bfc, hits$rid, fpath = fpath, ask = FALSE)
    lpath <- BiocFileCache::bfcpath(bfc, hits$rid)
  }

  # multiple entries found, clean up
  else {
    BiocFileCache::bfcremove(bfc, hits$rid)
    lpath <- BiocFileCache::bfcadd(bfc, rname, fpath)
  }

  return(lpath)
}


#' Cached URL path
#'
#' This function allows you to obtain a cached local path for a resource
#' identified by its name and a static URL. If caching is enabled (controlled by
#' the `use_cache` argument), it uses a `BiocFileCache` object to store and
#' retrieve the resource. If caching is disabled, the original file path is
#' returned.
#'
#' @param rname A character string representing the resource name.
#' @param fpath A character string with the static URL of the file.
#' @param use_cache A logical value indicating whether to use caching. If TRUE,
#'   the function caches the resource; if FALSE, it uses the original path.
#'
#' @return A character string representing either the cached local path (if
#'   caching is enabled) or the original file path (if caching is disabled).
#' @noRd
cached_url_path <- function(rname, fpath, use_cache) {
  if(use_cache) {
    cache <- cache_location()
    bfc <- BiocFileCache::BiocFileCache(cache, ask = FALSE)
    lpath <- cache_path(bfc, rname, fpath)
  } else {
    lpath <- fpath
  }
  return(lpath)
}


#' Remove all cache
#'
#' This function will remove all cached data used by `fenr`. The user will be
#' prompted for confirmation. Use only when you suspect the cache was corrupted.
#' Use with caution!
#'
#' @param ask Logical, whether to ask user for confirmation.
#' @return TRUE if successfully removed.
#' @export
#' @examples
#' \dontrun{
#' remove_cache()
#' }
remove_cache <- function(ask = TRUE) {
  cache <- cache_location()
  bfc <- BiocFileCache::BiocFileCache(cache, ask = FALSE)
  BiocFileCache::removebfc(bfc, ask = ask)
}

test_that("Cache location exists", {
  cache <- cache_location()
  expect_true(dir.exists(cache))
})

test_that("Clearing cache", {
  remove_cache(ask = FALSE)

  cache <- cache_location()
  bfc <- BiocFileCache::BiocFileCache(cache, ask = FALSE)
  cnt <- BiocFileCache::bfccount(bfc)

  expect_equal(cnt, 0)
})


fpath <- "https://raw.githubusercontent.com/bartongroup/fenr/main/DESCRIPTION"

# proceed with tests only if file is available (ingnore internet issues)

if(assert_url_path(fpath, on_error = "ignore")) {

  rname <- "test"
  expected_content <- readLines(fpath)

  cleanup_test <- function(bfc) {
    hits <- BiocFileCache::bfcquery(bfc, rname, field = "rname")
    if(nrow(hits) > 0)
      BiocFileCache::bfcremove(bfc, hits$rid)
  }

  test_that("No caching test", {
    re <- cached_url_path(rname, fpath, use_cache = FALSE)
    expect_equal(re, fpath)
  })


  test_that("Adding new entry to the cache", {
    cache <- cache_location()
    bfc <- BiocFileCache::BiocFileCache(cache, ask = FALSE)
    cleanup_test(bfc)

    re <- cached_url_path(rname, fpath, use_cache = TRUE)
    expect_true(file.exists(re))
    re_content <- readLines(re)
    expect_equal(re_content, expected_content)
  })


  test_that("Retrieving an entry to the cache", {
    cache <- cache_location()
    bfc <- BiocFileCache::BiocFileCache(cache, ask = FALSE)
    cleanup_test(bfc)
    lpath <- BiocFileCache::bfcadd(bfc, rname, fpath)

    re <- cached_url_path(rname, fpath, use_cache = TRUE)
    expect_true(file.exists(re))
    expect_equal(re, lpath)

    re_content <- readLines(re)
    expect_equal(re_content, expected_content)
  })


  test_that("Multiple entries behaviour", {
    cache <- cache_location()
    bfc <- BiocFileCache::BiocFileCache(cache, ask = FALSE)
    cleanup_test(bfc)
    lpath1 <- BiocFileCache::bfcadd(bfc, rname, fpath)
    lpath2 <- BiocFileCache::bfcadd(bfc, rname, fpath)

    re <- cached_url_path(rname, fpath, use_cache = TRUE)
    expect_true(file.exists(re))

    re_content <- readLines(re)
    expect_equal(re_content, expected_content)
  })

}

#' Create term data for interactive example
#'
#' @param de Differential expression results, use \code{yeast_de} data attached
#'   to this package.
#'
#' @return A list of objects containing functional terms for GO and Reactome.
#' @export
#' @examples
#' \dontrun{
#' data(yeast_de)
#' term_data <- fetch_terms_for_example(yeast_de)
#' }
fetch_terms_for_example <- function(de) {
  # All gene background
  all_genes <- de$gene_id

  # load GO terms
  message("Fetching GO data\n")
  go <- fetch_go(species = "sgd")
  go_data <- prepare_for_enrichment(go$terms, go$mapping, all_genes, feature_name = "gene_id")

  # load Reactome pathways
  message("Fetching Reactome data\n")
  re <- fetch_reactome("Saccharomyces cerevisiae")
  re_data <- prepare_for_enrichment(re$terms, re$mapping, all_genes, feature_name = "gene_id")

  # Put all functional term data in one structure; Shiny app will access
  # individual ontologies from this list
  list(
    go = go_data,
    re = re_data
  )
}


#' Volcano plot
#'
#' @param d Tibble with x, y and FDR
#' @param fdr_limit FDR limit below which point will be plotted black
#'
#' @return A ggplot object
#' @noRd
plot_volcano <- function(d, fdr_limit = 0.05) {
  # Binding variables from non-standard evaluation locally
  x <- y <- FDR <- NULL

  sres <- d |>
    dplyr::filter(FDR <= fdr_limit)
  g <- ggplot2::ggplot(d, ggplot2::aes(x, y)) +
    ggplot2::theme_classic() +
    ggplot2::theme(
      panel.grid = ggplot2::element_blank(),
      text = ggplot2::element_text(size = 18)
    ) +
    ggplot2::geom_point(size = 0.2) +
    ggplot2::geom_vline(xintercept = 0, colour = "grey70") +
    ggplot2::labs(x = expression(log[2]~FC), y = expression(-log[10]~P)) +
    ggplot2::scale_y_continuous(expand = ggplot2::expansion(mult = c(0, 0.03)))
  if (nrow(sres) > 0) {
    g <- g + ggplot2::geom_hline(yintercept = -log10(max(sres$PValue)), colour = "red",
                        linetype = "dashed", alpha = 0.2)
  }
  g
}

#' MA plot
#'
#' @param d Tibble with x, y and FDR
#' @param fdr_limit FDR limit below which point will be plotted black
#'
#' @return A ggplot object
#' @noRd
plot_ma <- function(d, fdr_limit = 0.05) {
  # Binding variables from non-standard evaluation locally
  x <- y <- NULL

  ggplot2::ggplot(d, ggplot2::aes(x, y)) +
    ggplot2::theme_classic() +
    ggplot2::theme(
      panel.grid = ggplot2::element_blank(),
      text = ggplot2::element_text(size = 18)
    ) +
    ggplot2::geom_point(data = d[d$FDR > fdr_limit, ], size = 0.1, colour = "grey50") +
    ggplot2::geom_point(data = d[d$FDR <= fdr_limit, ], size = 0.2, colour = "black") +
    ggplot2::geom_hline(yintercept = 0, colour = "grey70") +
    ggplot2::labs(x = expression(log[2]~CPM), y = expression(log[2]~FC))
}



#' Helper function for Shiny app
#'
#' @param de Differential expression results
#' @param input App input
#'
#' @return A tibble with x, y
#' @noRd
get_xy_data <- function(de, input) {
  logFC <- PValue <- logCPM <- NULL
  if (input$plot_type == "volcano") {
    xy_data <- de |>
      dplyr::mutate(x = logFC, y = -log10(PValue))
  } else if (input$plot_type == "ma") {
    xy_data <- de  |>
      dplyr::mutate(x = logCPM, y = logFC)
  }
  xy_data
}

#' Helper function for Shiny app
#'
#' @param de Differential expression results
#' @param term_data All functional term data
#' @param input App input
#' @param max_points Maximum number of points allowed to select
#'
#' @return A tibble with functional enrichment results
#' @noRd
enrichment_table <- function(de, term_data, input, max_points = 10000) {
  # Binding variables from non-standard evaluation locally
  p_adjust <- n_with_sel <- NULL

  xy_data <- get_xy_data(de, input)
  terms <- term_data[[input$ontology]]
  gene2symbol <- rlang::set_names(de$gene_symbol, de$gene_id)
  sel <- NULL
  fe <- NULL
  if (!is.null(input$plot_brush)) {
    brushed <- shiny::brushedPoints(xy_data, input$plot_brush)
    sel <- brushed$gene_id
    n <- length(sel)
    if (n > 0 && n <= max_points) {
      fe <- functional_enrichment(de$gene_id, sel, terms, feat2name = gene2symbol)
      if(!is.null(fe))
        fe <- dplyr::filter(fe, p_adjust < 0.05 & n_with_sel > 2)
    } else if (n > 0) {
      fe <- data.frame(Error = paste("only", max_points, "points can be selected."))
    }
  }
  fe
}

#' Helper function for Shiny app
#'
#' @param de Differential expression results
#' @param input App input
#'
#' @return Data for volcano or ma plot
#' @noRd
main_plot <- function(de, input) {
  xy_data <- get_xy_data(de, input)
  if (input$plot_type == "volcano") {
    plot_volcano(xy_data)
  } else if (input$plot_type == "ma") {
    plot_ma(xy_data)
  }
}


# Note for Bioconductor: Using "donttest" is problematic here. When running
# `devtools::check()`, it stalls at "checking examples with --run-donttest ...".
# I suspect it's attempting to launch the Shiny app, which fails. As a result,
# contrary to `BiocCheck()` guidelines, we've opted for "dontrun" in this instance.


#' Small Shiny app serving as example for fast enrichment
#'
#' @param de Differential expression results, \code{yeast_de} attached to this
#'   package can be used.
#' @param term_data A list of \code{fenr_terms} objects containing functional
#'   data for various ontologies. \code{fetch_terms_for_example} can be used to
#'   create this object.
#'
#' @return An interactive Shiny app
#' @importFrom assertthat assert_that
#' @export
#' @examples
#' \dontrun{
#' data(yeast_de)
#' term_data <- fetch_terms_for_example(yeast_de)
#'   enrichment_interactive(yeast_de, term_data)
#' }
enrichment_interactive <- function(de, term_data) {
  assert_that(is.data.frame(de))
  assert_that(is(term_data, "list"))

  req_cols <- c("gene_id", "gene_symbol", "logFC", "logCPM", "PValue", "FDR")
  if(!all(req_cols %in% colnames(de)))
    stop(paste0("The data frame needs to contain the following columns\n", paste(req_cols, collapse = ", ")))


  ui <- function() {
    shiny::shinyUI(shiny::fluidPage(
      shiny::tags$style("table{font-size: 11px; background-color: #EAF5FF}"),
      shiny::titlePanel("fenr example"),
      shiny::p("Select a group of genes in the plot to see their functional enrichment."),

      shiny::fluidRow(
        shiny::column(5,
          shiny::radioButtons(
            inputId = "plot_type",
            label = "Plot type:",
            choices = c("Volcano" = "volcano", "MA" = "ma"),
            inline = TRUE
          ),
          shiny::plotOutput(
            outputId = "main_plot",
            height = "480px",
            width = "100%",
            brush = "plot_brush",
            hover = "plot_hover"
          )
        ),
        shiny::column(7,
          shiny::radioButtons(
            inputId = "ontology",
            label = "Ontology:",
            choices = c("GO" = "go", "Reactome" = "re"),
            inline = TRUE
          ),
          shiny::div(style = "height: 480px; overflow-y: scroll", shiny::tableOutput("enrichment")),
        )
      )
    ))
  }

  server <- function(input, output, session) {
    # Prevents RStudio from crashing when Shiny window closed manually
    session$onSessionEnded(function() {
      shiny::stopApp()
    })

    output$enrichment <- shiny::renderTable({
      enrichment_table(de, term_data, input)
    })

    output$main_plot <- shiny::renderPlot({
      main_plot(de, input)
    })
  }

  # Run the application
  shiny::shinyApp(ui = ui, server = server)
}

#' Prepare Term Data for Enrichment Analysis
#'
#' Process term data downloaded with the \code{fetch_*} functions, preparing it
#' for fast enrichment analysis using \code{functional_enrichment}.
#'
#' @details
#'
#' This function takes two tibbles containing functional term information
#' (\code{terms}) and feature mapping (\code{mapping}), and converts them into
#' an object required by \code{functional_enrichment} for efficient analysis.
#' Terms and mapping can be generated with the database access functions
#' included in this package, such as \code{fetch_reactome} or
#' \code{fetch_go_from_go}.
#'
#' @param terms A tibble with at least two columns: \code{term_id} and
#'   \code{term_name}. This tibble contains information about functional term
#'   names and descriptions.
#' @param mapping A tibble with at least two columns, containing the mapping
#'   between functional terms and features. One column must be named
#'   \code{term_id}, while the other column should have a name specified by the
#'   \code{feature_name} argument. For example, if \code{mapping} contains
#'   columns \code{term_id}, \code{accession_number}, and \code{gene_symbol},
#'   setting \code{feature_name = "gene_symbol"} indicates that gene symbols
#'   will be used for enrichment analysis.
#' @param all_features A vector with all feature IDs used as the background for
#'   enrichment. If not specified, all features found in \code{mapping} will be
#'   used, resulting in a larger object size.
#' @param feature_name The name of the column in the \code{mapping} tibble to be
#'   used as the feature identifier. For example, if \code{mapping} contains
#'   columns \code{term_id}, \code{accession_number}, and \code{gene_symbol},
#'   setting \code{feature_name = "gene_symbol"} indicates that gene symbols
#'   will be used for enrichment analysis.
#'
#' @return An object of class \code{fenr_terms} required by
#'   \code{functional_enrichment}.
#' @importFrom assertthat assert_that
#' @export
#' @examples
#' \dontrun{
#' data(exmpl_all)
#' go <- fetch_go(species = "sgd")
#' go_terms <- prepare_for_enrichment(go$terms, go$mapping, exmpl_all,
#'                                    feature_name = "gene_symbol")
#' }
prepare_for_enrichment <- function(terms, mapping, all_features = NULL,
                                   feature_name = "gene_id") {
  # Binding variables from non-standard evaluation locally
  feature_id <- term_id <- NULL

  # Argument checks
  assert_that(is.data.frame(terms) || tibble::is_tibble(terms),
              msg = "'terms' must be a data frame or tibble.")

  assert_that(is.data.frame(mapping) || tibble::is_tibble(mapping),
              msg = "'mapping' must be a data frame or tibble.")

  assert_that(is.null(all_features) || is.vector(all_features),
              msg = "'all_features' must be a vector or NULL.")

  assert_that(is.character(feature_name) && length(feature_name) == 1,
              msg = "'feature_name' must be a single string.")

  # Check terms
  assert_that(all(c("term_id", "term_name") %in% colnames(terms)),
              msg = "Column names in 'terms' should be 'term_id' and 'term_name'.")

  assert_that(anyDuplicated(terms$term_id) == 0,
              msg = "Duplicated term_id detected in 'terms'.")

  # Check mapping
  assert_that("term_id" %in% colnames(mapping),
              msg = "'mapping' should contain a column named 'term_id'.")

  # Check for feature name
  assert_that(feature_name %in% colnames(mapping),
              msg = paste0(feature_name, " column not found in mapping table.
                           Check 'feature_name' argument."))

  # Replace empty all_features with everything from mapping
  map_features <- mapping[[feature_name]] |>
    unique()
  if (is.null(all_features)) {
    all_features <- map_features
  } else {
    # Check if mapping is contained in all features
    if (length(intersect(all_features, map_features)) == 0)
      stop("No overlap between 'all_features' and features found in 'mapping'.
           Did you provide correct 'all_features'?")
  }

  # Check for missing term descriptions
  mis_term <- setdiff(mapping$term_id, terms$term_id)
  if (length(mis_term) > 0) {
    dummy <- tibble::tibble(
      term_id = mis_term,
      term_name = rep(NA_character_, length(mis_term))
    )
    terms <- dplyr::bind_rows(terms, dummy)
  }

  # Hash to select term name
  term2name <- new.env(hash = TRUE)
  for (i in seq_len(nrow(terms))) {
    r <- terms[i, ]
    term2name[[r$term_id]] <- r$term_name
  }

  # feature-term tibble
  feature_term <- mapping |>
    dplyr::rename(feature_id = !!feature_name) |>
    dplyr::filter(feature_id %in% all_features) |>
    dplyr::select(feature_id, term_id) |>
    dplyr::distinct()

  # Feature to terms hash
  f2t <- feature_term |>
    dplyr::group_by(feature_id) |>
    dplyr::summarise(terms = list(term_id)) |>
    tibble::deframe()
  feature2term <- new.env(hash = TRUE)
  for(feat in names(f2t))
    feature2term[[feat]] <- f2t[[feat]]

  # Term to feature hash
  t2f <- feature_term |>
    dplyr::group_by(term_id) |>
    dplyr::summarise(features = list(feature_id)) |>
    tibble::deframe()
  term2feature <- new.env(hash = TRUE)
  for(term in names(t2f))
    term2feature[[term]] <- t2f[[term]]

  list(
    term2name = term2name,
    term2feature = term2feature,
    feature2term = feature2term
  ) |>
    structure(class = "fenr_terms")
}



#' Fast Functional Enrichment
#'
#' Perform fast functional enrichment analysis based on the hypergeometric
#' distribution. Designed for use in interactive applications.
#'
#' @details This function carries out functional enrichment analysis on a
#'   selection of features (e.g., differentially expressed genes) using the
#'   hypergeometric probability distribution (Fisher's exact test). Features can
#'   be genes, proteins, etc. The \code{term_data} object contains functional
#'   term information and feature-term mapping.
#'
#' @param feat_all A character vector with all feature identifiers, serving as
#'   the background for enrichment.
#' @param feat_sel A character vector with feature identifiers in the selection.
#' @param term_data An object of class \code{fenr_terms}, created by
#'   \code{prepare_for_enrichment}.
#' @param feat2name An optional named list to convert feature IDs into feature
#'   names.
#'
#' @return A tibble with enrichment results, providing the following information
#'   for each term:
#'   \itemize{
#'     \item{\code{N_with} - number of features with this term among all features}
#'     \item{\code{n_with_sel} - number of features with this term in the selection}
#'     \item{\code{n_expect} - expected number of features with this term in the selection,
#'       under the null hypothesis that terms are mapped to features randomly}
#'     \item{\code{enrichment} - ratio of n_with_sel / n_expect}
#'     \item{\code{odds_ratio} - odds ratio for enrichment; is infinite when all
#'       features with the given term are in the selection}
#'     \item{\code{p_value} - p-value from a single hypergeometric test}
#'     \item{\code{p_adjust} - p-value adjusted for multiple tests using the
#'       Benjamini-Hochberg approach}
#'   }.
#'
#' @importFrom assertthat assert_that
#' @importFrom methods is
#' @export
#' @examples
#' \dontrun{
#' data(exmpl_all, exmpl_sel)
#' go <- fetch_go(species = "sgd")
#' go_terms <- prepare_for_enrichment(go$terms, go$mapping, exmpl_all, feature_name = "gene_symbol")
#' enr <- functional_enrichment(exmpl_all, exmpl_sel, go_terms)
#' }
functional_enrichment <- function(feat_all, feat_sel, term_data, feat2name = NULL) {

  # Binding variables from non-standard evaluation locally
  N_with <- n_with_sel <- n_expect <- enrichment <- odds_ratio <- NULL
  desc <- p_value <- p_adjust <- NULL

  # Check for character vectors
  assert_that(is.character(feat_all))
  assert_that(is.character(feat_sel))
  assert_that(length(feat_all) > 1)
  assert_that(length(feat_sel) > 1)

  # Check term_data class
  assert_that(is(term_data, "fenr_terms"))

  # If no overlap between selection and all, return NULL
  if(!any(feat_sel %in% feat_all))
    return(NULL)

  # all terms present in the selection
  our_terms <- feat_sel |>
    purrr::map(~term_data$feature2term[[.x]]) |>
    unlist() |>
    unique()

  # number of features in selection
  N_sel <- length(feat_sel)
  # total number of features
  N_tot <- length(feat_all)

  res <- purrr::map_dfr(our_terms, function(term_id) {
    # all features with the term
    # term_data$term2feature is a hash environment
    tfeats <- term_data$term2feature[[term_id]]
    # necessary if term data contain features not present in feat_all
    tfeats <- tfeats[tfeats %in% feat_all]

    # features from selection with the term
    # this is faster than intersect(tfeats, feat_sel)
    tfeats_sel <- tfeats[tfeats %in% feat_sel]

    N_with <- length(tfeats)
    N_without <- N_tot - N_with

    # building contingency table
    n_with_sel <- length(tfeats_sel)
    n_without_sel <- N_sel - n_with_sel
    n_with_nsel <- N_with - n_with_sel
    n_without_nsel <- N_tot - (n_with_sel + n_without_sel + n_with_nsel)

    # contingency table is
    #               | In selection  | Not in selection
    #-------------------------------------------------
    #  With term    | n_with_sel    | n_with_nsel
    #  Without term | n_without_sel | n_without_nsel

    if (n_with_sel < 2) return(NULL)

    # Expected number of features in selection, if random
    n_expect <- N_with * N_sel / N_tot

    # Odds ratio
    odds_ratio <- (n_with_sel / n_without_sel) / (n_with_nsel / n_without_nsel)

    # Hypergeometric function much faster than fisher.test
    p <- 1 - stats::phyper(n_with_sel - 1, N_with, N_without, N_sel)

    # Convert feature IDs to feature names;
    if (!is.null(feat2name))
      tfeats_sel <- feat2name[tfeats_sel] |> unname()

    term_name <- term_data$term2name[[term_id]]
    # returns NAs if no term found
    if (is.null(term_name)) term_name <- NA_character_

    # constructing a tibble in every iteration is more time expensive than `c`,
    # even with overhead of converting types afterwards. Not elegant, but fast.
    c(
      term_id = term_id,
      term_name = term_name,
      N_with = N_with,
      n_with_sel = n_with_sel,
      n_expect = n_expect,
      enrichment = n_with_sel / n_expect,
      odds_ratio = odds_ratio,
      ids = paste(tfeats_sel, collapse = ", "),
      p_value = p
    )
  })

  # Drawback - if all selections below minimum, res is tibble 0 x 0, need to
  # catch it
  if (nrow(res) == 0) {
    res <- NULL
  } else {
    res <- res |>
      dplyr::mutate(
        dplyr::across(c(N_with, n_with_sel), as.integer),
        dplyr::across(c(n_expect, enrichment, odds_ratio, p_value), as.numeric),
        p_adjust = stats::p.adjust(p_value, method = "BH"),
        dplyr::across(c(enrichment, odds_ratio, p_value, p_adjust), ~signif(.x, 3)),
        n_expect = round(n_expect, 2)
      ) |>
      dplyr::arrange(desc(odds_ratio))
  }
  res
}

# Gene Annotation File column names. GAF is used by GO and Reactome.
GAF_COLUMNS <- c(
  "db",
  "db_id",
  "symbol",
  "qualifier",
  "go_term",
  "db_ref",
  "evidence",
  "from",
  "aspect",
  "db_object_name",
  "db_object_synonym",
  "db_object_type",
  "taxon",
  "date",
  "assigned_by",
  "annotation_extension",
  "form_id"
)

GAF_TYPES <- rep("c", length(GAF_COLUMNS)) |>
  stringr::str_c(collapse = "")


#' Check if URL with a path is responding OK
#'
#' Stops with error message if the path is not accessible or server not
#' responding.
#'
#' @param url_path Full URL with a path, e.g. `https://reactome.org/download/current/ReactomePathways.txt`.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#' @param timeout Timeout limit in seconds.
#'
#' @importFrom assertthat assert_that is.string
#' @return TRUE if assertion passed
#' @noRd
assert_url_path <- function(url_path, on_error = c("stop", "warn", "ignore"), timeout = 30) {
  on_error <- match.arg(on_error)
  assert_that(is.string(url_path))

  resp <- http_request(url_path, "", timeout = timeout)

  if(resp$is_error) {
    catch_error(stringr::str_glue("{url_path}"), resp, on_error)
    return(FALSE)
  }
  return(TRUE)
}



#' Check if species are valid
#'
#' @param species A string, species designation for a given database
#' @param fetch_fun A string, name of the function to retrieve available
#'   species, must return a tibble with a column \code{designation}.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @importFrom assertthat assert_that is.string
#' @return A tibble with valid species - a response from \code{fetch_fun}
#' @noRd
assert_species <- function(species, fetch_fun, on_error) {
  assert_that(is.string(species))
  assert_that(is.string(fetch_fun))
  f <- match.fun(fetch_fun)
  valid_species <- f(on_error = on_error)
  if(is.null(valid_species))
    return(NULL)
  assert_that(species %in% valid_species$designation,
    msg = stringr::str_glue("Invalid species {species}. Use {fetch_fun}() to find all available species.")
  )
  valid_species
}


#' Return a field for a given species
#'
#' Used to match species designation with a species ID.
#'
#' @param species Species designation
#' @param fetch_fun  A string with name of a function to retrieve available
#'   species, must return a tibble with a column \code{designation}.
#' @param col_name Column name in the tibble returned by \code{fetch_fun} to
#'   extract, e.g. \code{tax_id}
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A value extracted from column \code{col_name} at row where
#'   \code{designation} = \code{species}.
#' @noRd
match_species <- function(species, fetch_fun, col_name, on_error) {
  designation <- NULL

  sp <- assert_species(species, fetch_fun, on_error)
  if(is.null(sp))
    return(NULL)
  sp |>
    dplyr::filter(designation == species) |>
    dplyr::pull(get(col_name)) |>
    unique()
}


#' Get features annotated with a given term.
#'
#' @param term_data An object class \code{fenr_terms}, created by
#'   \code{prepare_for_enrichment}.
#' @param term_id A string with a functional term ID.
#'
#' @return A character vector containing feature IDs annotated with the term ID.
#'
#' @importFrom methods is
#' @importFrom assertthat assert_that is.string
#' @export
#' @examples
#' \dontrun{
#' go_data <- fetch_go(species = "sgd")
#' go_terms <- prepare_for_enrichment(go_data$terms, go_data$mapping, feature = "gene_symbol")
#' feats <- get_term_features(go_terms, "GO:0000001")
#' }
get_term_features <- function(term_data, term_id) {
  # Check term_data class
  assert_that(is(term_data, "fenr_terms"))
  assert_that(is.string(term_id))

  term_data$term2feature[[term_id]]
}


#' Get terms annotating a given feature.
#'
#' @param term_data An object class \code{fenr_terms}, created by
#'   \code{prepare_for_enrichment}.
#' @param feature_id A string with a feature ID
#'
#' @return A character vector containing functional term IDs annotating given feature.
#' @importFrom assertthat assert_that is.string
#' @importFrom methods is
#' @export
#' @examples
#' \dontrun{
#' go_data <- fetch_go(species = "sgd")
#' go_terms <- prepare_for_enrichment(go_data$terms, go_data$mapping, feature = "gene_symbol")
#' trms <- get_feature_terms(go_terms, "GEM1")
#' }
get_feature_terms <- function(term_data, feature_id) {
  # Check term_data class
  assert_that(is(term_data, "fenr_terms"))
  assert_that(is.string(feature_id))

  term_data$feature2term[[feature_id]]
}



#' Test correctness of data structure returned by fetch_*
#'
#' @param re The data structure returned by fetch_*
#' @noRd
test_fetched_structure <- function(re) {
  testthat::expect_is(re, "list")
  testthat::expect_length(re, 2)
  testthat::expect_named(re)
  testthat::expect_equal(names(re), c("terms", "mapping"))
}


#' Test if returned terms tibble is consistent with expected selection
#'
#' @param returned A tibble, the terms data returned by  fetch_*.
#' @param expected A tibble, a selection of expected terms data.
#' @noRd
test_terms <- function(returned, expected) {
  # Binding variables from non-standard evaluation locally
  term_id <- term_name <- NULL

  testthat::expect_is(returned, "tbl")
  testthat::expect_contains(names(returned), c("term_id", "term_name"))

  returned <- returned |> dplyr::select(term_id, term_name)

  merged <- expected |>
    dplyr::left_join(returned, by = c("term_id", "term_name")) |>
    tidyr::drop_na()
  testthat::expect_equal(nrow(expected), nrow(merged))
}

#' Test if returned mapping is consistent with expected selection
#'
#' @param returned A tibble, the mapping returned by  fetch_*.
#' @param expected A tibble, a selection of expected mapping.
#' @param feature_id A character string, the name of the column containing the feature
#'   ID.
#' @noRd
test_mapping <- function(returned, expected, feature_id) {
  # Binding variables from non-standard evaluation locally
  term_id <- NULL

  testthat::expect_is(returned, "tbl")
  testthat::expect_contains(names(returned), c("term_id", feature_id))

  returned <- returned |> dplyr::select(term_id, tidyselect::all_of(feature_id))

  merged <- expected |>
    dplyr::left_join(returned, by = c("term_id", feature_id)) |>
    dplyr::distinct() |>
    tidyr::drop_na()
  testthat::expect_equal(nrow(expected), nrow(merged))
}

#' Catch and Handle Errors Based on Specified Action
#'
#' This function handles errors by either stopping execution or issuing a
#' warning, depending on the specified action. It takes a message and an action
#' choice (`"stop"` or `"warn"`) as inputs.
#'
#' @param server Name of the server.
#' @param resp Response from \code{http_request()}. A list with "status" and
#'   "desc" strings.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return No return value if `on_error` is "stop". If `on_error` is "warn" or
#'   "ignore", the function returns `NULL` after issuing a warning.
#' @noRd
catch_error <- function(server, resp, on_error = c("stop", "warn", "ignore")) {
  on_error <- match.arg(on_error)

  st <- stringr::str_glue("Cannot access {server}. {resp$status}: {resp$description}.")

  error_response(st, on_error)
}

#' Error Response Handler
#'
#' This function handles errors by either stopping execution or issuing a
#' warning based on the specified action. It is designed to provide a
#' standardized way to handle error messages within functions.
#'
#' @param msg A character string specifying the error message to be displayed.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return No return value if `on_error` is "stop". If `on_error` is "warn" or
#'   "ignore", the function returns `NULL` after issuing a warning.
#' @noRd
error_response <- function(msg, on_error = c("stop", "warn", "ignore")) {
  on_error <- match.arg(on_error)

  if(on_error == "stop") {
    stop(msg)
  } else if (on_error == "warn") {
    warning(msg, "\nNULL returned.", call. = FALSE)
    return(NULL)
  } else {
    return(NULL)
  }
}

#' Perform a HTTP request using httr2
#'
#' This function sends an HTTP request to a specified API endpoint and returns
#' various details about the response. It is designed to work with the httr2
#' package.
#'
#' @param base_url A string specifying the base URL of the API.
#' @param path A string specifying the path of the specific API endpoint.
#' @param parameters An optional named list of query parameters to be included
#'   in the request.
#' @param timeout Timeout limit in seconds.
#'
#' @return A list containing the following elements:
#' \itemize{
#'   \item{response}{The complete response object returned by the API.}
#'   \item{is_error}{Logical value indicating if the response was an error.}
#'   \item{status}{The HTTP status code of the response.}
#'   \item{description}{The description of the HTTP status.}
#' }
#'
#' @details The function constructs a request using the base URL and the path.
#'   If provided, query parameters are appended to the request. The function
#'   then performs the request and checks for errors. It returns a list
#'   containing the response, error status, HTTP status code, and the
#'   description of the status. In case of a time-out, when the server does not
#'   return any code, a list with NULL response, and fixed status and
#'   description are returned.
#' @noRd
http_request <- function(base_url, path, parameters = NULL, timeout = 30) {
  req <- httr2::request(base_url) |>
    httr2::req_url_path_append(path) |>
    httr2::req_timeout(timeout)

  if(!is.null(parameters)) {
    req <- req |>
      httr2::req_url_query(!!!parameters)
  }

  resp <- tryCatch(
    {req |>
        httr2::req_error(is_error = ~FALSE) |>
        httr2::req_perform()
    },
    error = function(err) NULL
  )

  if(is.null(resp)) {
    ans <- list(
      response = NULL,
      is_error = TRUE,
      status = "unknown",
      description = "Unknown error (probably timeout)"
    )
  } else {
    ans <- list(
      response = resp,
      is_error = httr2::resp_is_error(resp),
      status = httr2::resp_status(resp),
      description = httr2::resp_status_desc(resp)
    )
  }

  ans
}


#' Test a function for correct response when server is non-responsive.
#'
#' @details When the server is not responsive (using
#'   httr2::with_mocked_responses), the function should throw an error, if
#'   `on_error = "stop"` and throw a warning and return NULL, it `on_error =
#'   "warn"`.
#'
#'
#' @param fun A function to test
#' @param ... Parameters to pass to the function
#'
#' @return Test result
#' @noRd
test_unresponsive_server <- function(fun, ...) {
  testthat::expect_error(fun(..., on_error = "stop"))
  testthat::expect_warning({res <- fun(..., on_error = "warn")})
  testthat::expect_null(res)
}


#' A mock function used with httr2::with_mocked_responses() for testing an
#' unresponsive server behaviour.
#'
#' @param req A httr2 request
#'
#' @return Mocked response with error 500.
#' @noRd
mocked_500 <- function(req) {
  httr2::response(status_code = 500)
}

library(testthat)

# Set 100 features
N <- 100
features_all <- sprintf("gene_%03d", seq_len(N))
symbols_all <- sprintf("symbol_%03d", seq_len(N))
feat2name <- purrr::set_names(symbols_all, features_all)

# Three terms with 50, 10 and 3 features
term_ids <- c("term_1", "term_2", "term_3")
term_names <- c("name_1", "name_2", "name_3")
term_sizes <- c(50, 10, 3)

# Prepare data for enrichment. The last term's name is intentionally missing.
term2name <- new.env(hash = TRUE)
for (i in seq_len(length(term_ids) - 1)) {
  term2name[[term_ids[i]]] <- term_names[i]
}

# random selection of features for terms
set.seed(666)
mapping <- purrr::map2_dfr(term_ids, term_sizes, function(tid, n) {
  tibble::tibble(
    term_id = tid,
    feature_id = sample(features_all, n)
  )
})

# Feature to terms hash
feature2term <- new.env(hash = TRUE)
features <- unique(mapping$feature_id)
for(feat in features) {
  terms <- mapping |>
    dplyr::filter(feature_id == feat) |>
    dplyr::pull(term_id)
  feature2term[[feat]] <- terms
}

# Term to feature hash
term2feature <- new.env(hash = TRUE)
terms <- unique(mapping$term_id)
for(term in terms) {
  features <- mapping |>
    dplyr::filter(term_id == term) |>
    dplyr::pull(feature_id)
  term2feature[[term]] <- features
}

# final structure required by functional_enrichment
term_data <- list(
  term2name = term2name,
  feature2term = feature2term,
  term2feature = term2feature
) |>
  structure(class = "fenr_terms")

term_stats <- function(tid, N_sel, n_with_sel) {
  N_with <- length(term2feature[[tid]])
  n_without_sel <- N_sel - n_with_sel
  n_expect <- N_with * N_sel / N
  enrichment <- n_with_sel / n_expect

  n_with_nsel <- N_with - n_with_sel
  n_without_nsel <- N - (n_with_sel + n_without_sel + n_with_nsel)
  cont_tab <- rbind(
    c(n_with_sel, n_with_nsel),
    c(n_without_sel, n_without_nsel)
  )

  # in the test we use fisher.test function as opposed to phyper in the code
  # p <- 1 - stats::phyper(n_with_sel - 1, N_with, N_without, N_sel)
  p <- fisher.test(cont_tab, alternative = "greater")$p.value

  sel_with <- term2feature[[tid]][seq_len(n_with_sel)]
  if (n_without_sel > 0) {
    sel_without <- setdiff(features_all, term2feature[[tid]])[seq_len(n_without_sel)]
  } else {
    sel_without <- character(0)
  }
  sel <- c(sel_with, sel_without)

  list(
    N_with = N_with,
    n_with_sel = n_with_sel,
    n_expect = n_expect,
    enrichment = enrichment |> signif(3),
    p = p |> signif(3),
    sel = sel
  )
}

# Test selection of 3 genes
set.seed(42)
test_sel <- mapping |>
  dplyr::group_by(term_id) |>
  dplyr::sample_n(3) |>
  dplyr::pull(feature_id)


#######################################################

test_that("Incorrect term_data class", {
  td <- term_data
  class(td) <- "wobble"

  expect_error(
    functional_enrichment(features_all, features_all[seq_len(10)], td)
  )
})


test_that("Features must be character vectors with length > 1", {
  expect_error(functional_enrichment(c(1, 2, 3, 4), c(1, 2), term_data))
  expect_error(functional_enrichment(features_all, features_all[1], term_data))
  expect_error(functional_enrichment(features_all, character(0), term_data))
  expect_error(functional_enrichment(features_all[1], features_all[1], term_data))
  expect_error(functional_enrichment(character(0), character(0), term_data))
})

test_that("Correct output", {
  enr <- functional_enrichment(features_all, test_sel, term_data)

  expect_true(is.data.frame(enr))
  expect_equal(nrow(enr), length(term_ids))
  expect_equal(ncol(enr), 10)
})

test_that("Gene names translated", {
  enr <- functional_enrichment(features_all, test_sel, term_data, feat2name)
  symbols <- enr |>
    tidyr::separate_longer_delim(ids, delim = ", ") |>
    dplyr::pull(ids) |>
    unique()
  expect_contains(symbols_all, symbols)
})


test_that("Missing term name replaced with NA", {
  enr <- functional_enrichment(features_all, test_sel, term_data)
  misname <- enr |>
    dplyr::filter(term_id == "term_3") |>
    dplyr::pull(term_name)
  expect_true(is.na(misname))
})


test_that("n_with_sel < 2 should return NULL", {
  set.seed(42)
  # genes for term 1
  feat_1 <- mapping |>
    dplyr::filter(term_id == term_ids[1]) |>
    dplyr::pull(feature_id)
  # genes for term 2
  feat_2 <- mapping |>
    dplyr::filter(term_id == term_ids[2]) |>
    dplyr::pull(feature_id)
  # find genes exclusive to selections
  sel_1 <- setdiff(feat_1, feat_2)
  sel_2 <- setdiff(feat_2, feat_1)

  # One gene from feat_1, 3 genes from feat_2
  sel <- c(sample(sel_1, 1), sample(sel_2, 3))
  enr <- functional_enrichment(features_all, sel, term_data)
  # Term 1 should not appear in the output
  expect_false(term_ids[1] %in% enr$term_id)

  # One gene from feat_1, 1 gene from feat_2
  sel <- c(sample(sel_1, 1), sample(sel_2, 1))
  enr <- functional_enrichment(features_all, sel, term_data)
  # All searches have n_with_sel < 2, expect NULL
  expect_null(enr)
})


test_that("Enrichment of 1", {
  tid <- term_ids[1]
  N_sel <- 20
  n_with_sel <- 10

  res <- term_stats(tid, N_sel, n_with_sel)
  enr <- functional_enrichment(features_all, res$sel, term_data) |>
    dplyr::filter(term_id == tid)

  expect_equal(
    c(res$N_with, res$n_with_sel, res$n_expect, res$enrichment, res$p),
    c(enr$N_with, enr$n_with_sel, enr$n_expect, enr$enrichment, enr$p_value)
  )
})


test_that("All selected features in term", {
  tid <- term_ids[2]
  N_sel <- term_sizes[2]
  n_with_sel <- term_sizes[2]

  res <- term_stats(tid, N_sel, n_with_sel)
  enr <- functional_enrichment(features_all, res$sel, term_data) |>
    dplyr::filter(term_id == tid)

  expect_equal(
    c(res$N_with, res$n_with_sel, res$n_expect, res$enrichment, res$p),
    c(enr$N_with, enr$n_with_sel, enr$n_expect, enr$enrichment, enr$p_value)
  )
})


test_that("No selected features in term", {
  tid <- "term_3"
  N_sel <- 20
  n_with_sel <- 0

  res <- term_stats(tid, N_sel, n_with_sel)
  enr <- functional_enrichment(features_all, res$sel, term_data) |>
    dplyr::filter(term_id == tid)

  expect_equal(nrow(enr), 0)
})


test_that("Selection is all features", {
  enr <- functional_enrichment(features_all, features_all, term_data)

  n <- nrow(enr)
  expect_equal(enr$enrichment, rep(1, n))
  expect_equal(enr$p_value, rep(1, n))
})


test_that("No match between selection and all features", {
  expect_null(
    functional_enrichment(features_all, c("a", "b", "c"), term_data)
  )
})

library(testthat)


# Set 100 features
N <- 100
features_all <- sprintf("gene_%03d", seq_len(N))

# Three terms with 50, 10 and 3 features
term_ids <- c("term_1", "term_2", "term_3")
term_names <- c("name_1", "name_2", "name_3")
term_sizes <- c(50, 10, 3)


terms <- tibble::tibble(
  term_id = term_ids,
  term_name = term_names
)

# random selection of features for terms
set.seed(666)
mapping <- purrr::map2_dfr(term_ids, term_sizes, function(tid, n) {
  tibble::tibble(
    term_id = tid,
    feature_id = sample(features_all, n)
  )
}) |>
  dplyr::add_row(term_id = "missing_term", feature_id = "gene_000")

all_terms <- mapping$term_id |> unique()
all_features <- mapping$feature_id |> unique()


test_that("Expected correct output", {
  td <- prepare_for_enrichment(terms, mapping, feature_name = "feature_id")

  # Correct output
  expect_equal(length(td), 3)

  # Check class
  expect_s3_class(td, "fenr_terms")

  # Check term names
  expect_equal(length(td$term2name), length(all_terms))
  for(i in seq_along(terms$term_id)) {
    r <- terms[i, ]
    expect_equal(sort(r$term_name), sort(td$term2name[[r$term_id]]))
  }

  # Check for a missing term
  expect_contains(names(td$term2name), "missing_term")

  # Check term-feature hash
  term_ids <- mapping$term_id |> unique()
  chk1 <- term_ids |>
    purrr::map(function(trm) {
      expected <- mapping |>
        dplyr::filter(term_id == trm) |>
        dplyr::pull(feature_id) |>
        sort()
      returned <- td$term2feature[[trm]] |>
        sort()
      returned_fun <- get_term_features(td, trm) |>
        sort()
      expect_equal(expected, returned)
      expect_equal(expected, returned_fun)
    })

  # Check feature-term hash
  expect_equal(length(td$feature2term), length(all_features))
  chk2 <- all_features |>
    purrr::map(function(feat) {
      expected <- mapping |>
        dplyr::filter(feature_id == feat) |>
        dplyr::pull(term_id) |>
        sort()
      returned <- td$feature2term[[feat]] |>
        sort()
      returned_fun <- get_feature_terms(td, feat) |>
        sort()
      expect_equal(expected, returned)
      expect_equal(expected, returned_fun)
    })
})


test_that("Duplicate term-gene pairs are removed", {
  extra_row <- mapping[1, ]
  mapping_dup <- mapping |>
    dplyr::add_row(extra_row)
  td <- prepare_for_enrichment(terms, mapping_dup, feature_name = "feature_id")
  returned <- td$term2feature[[extra_row$term_id]] |>
    sort()
  expected <- mapping |>
    dplyr::filter(term_id == extra_row$term_id) |>
    dplyr::pull(feature_id) |>
    sort()
  expect_equal(expected, returned)
})


test_that("Wrong columns in terms", {
  trms <- terms |>
    dplyr::rename(gobble = term_id)

  expect_error(
    prepare_for_enrichment(trms, mapping, feature_name = "feature_id")
  )
})


test_that("Missing column in mapping", {
  mp <- mapping |>
    dplyr::rename(mobble = term_id)

  expect_error(
    prepare_for_enrichment(terms, mp, feature_name = "feature_id")
  )
})


test_that("Wrong feature name", {
  expect_error(
    prepare_for_enrichment(terms, mapping, feature_name = "trouble")
  )
})


test_that("No overlap between all_features and mapping", {
  expect_error(
    prepare_for_enrichment(terms, mapping, all_features = c("a", "b", "c"),  feature_name = "feature_id")
  )
})

#' @description
#'
#' This R package provides a fast and efficient method for functional enrichment
#' analysis, optimized for speed and designed for use in interactive
#' applications, such as *Shiny* apps.
#'
#' To learn more about fenr, start with the vignette: vignette("fenr").
#'
#' @keywords internal
#' @return Package fenr.
"_PACKAGE"

#' KEGG server base URL
#'
#' @return A string with URL. This can be changed by options(KEGG_BASE_URL =
#'   "different/url").
#' @noRd
get_kegg_url <- function() {
  getOption("KEGG_BASE_URL", "https://rest.kegg.jp")
}

#' Find all species available from KEGG
#'
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A tibble, in which column \code{designation} contains species
#'   designations used in function \code{fetch_kegg}.
#' @export
#' @examples
#' spe <- fetch_kegg_species(on_error = "warn")
fetch_kegg_species <- function(on_error = c("stop", "warn", "ignore")) {
  on_error <- match.arg(on_error)

  resp <- http_request(get_kegg_url(), "list/organism")
  if(resp$is_error)
    return(catch_error("KEGG", resp, on_error))

  st <- httr2::resp_body_string(resp$response)
  readr::read_tsv(I(st), col_names = c("id", "designation", "species", "phylogeny"), show_col_types = FALSE)
}


#' Download pathway data from KEGG
#'
#' @param species A valid species designation used by KEGG.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A tibble with columns \code{gene_id} and \code{term_id}.
#' @noRd
fetch_kegg_pathways <- function(species, on_error) {
  # Binding variables from non-standard evaluation locally
  term_id <- NULL

  resp <- http_request(get_kegg_url(), stringr::str_glue("list/pathway/{species}"))
  if(resp$is_error)
    return(catch_error("KEGG", resp, on_error))

  st <- httr2::resp_body_string(resp$response)
  readr::read_tsv(I(st), col_names = c("term_id", "term_name"), show_col_types = FALSE) |>
    dplyr::mutate(term_id = stringr::str_remove(term_id, "path:"))
}


#' Parse flat file returned from KEGG GET API
#'
#' @param s A string containing flat file result
#'
#' @return A tibble with \code{gene_is}, \code{gene_symbol} and \code{term_id}.
#' @noRd
parse_kegg_genes <- function(s) {
  # Binding variables from non-standard evaluation locally
  data <- gene_symbol <- gene_id <- NULL

  entries <- stringr::str_split(s, "///") |> unlist()
  entries <- entries[-length(entries)] # there is /// at the end
  purrr::map(entries, function(entry) {
    d <- stringr::str_split(entry, "\n") |>
      unlist()
    n <- length(d)
    i <- 1

    # find ENTRY key
    while(i <= n & !(stringr::str_detect(d[i], "^ENTRY")))
      i <- i + 1
    paths <- stringr::str_split(d[i], "\\s+") |> unlist()
    pathway <- paths[2]

    # find GENE key
    i <- 1
    while(i <= n & !(stringr::str_detect(d[i], "^GENE")))
      i <- i + 1
    # no GENE key found
    if(i > n)
      return(tibble::tibble(
        gene_id = character(0),
        gene_symbol = character(0)
      ))

    # extract genes
    genes <- stringr::str_remove(d[i], "GENE\\s+")
    i <- i + 1
    while(i <= n & stringr::str_detect(d[i], "^\\s+")) {
      genes <- c(genes, stringr::str_remove(d[i], "^\\s+"))
      i <- i + 1
    }

    purrr::map(genes, function(gene) {
      # First element - gene ID, second - gene symbol, if contains semicolon
      v <- stringr::str_split_1(gene, "\\s+")
      tibble::tibble(
        gene_id = v[1],
        gene_symbol = ifelse(
          stringr::str_detect(v[2], ";"),
          stringr::str_remove(v[2], ";.*$"),
          v[1]
        ),
        term_id = pathway
      )
    }) |> 
      purrr::list_rbind()
  }) |>
    purrr::list_rbind()
}


#' Download term - gene mapping from KEGG
#'
#' @param pathways A character vector with KEGG pathways
#' @param batch_size Number of pathways sent to KEGG database in one query. The
#'   maximum allowed is 10.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#' @importFrom assertthat assert_that
#' @return A tibble with columns \code{gene_id} and \code{term_id}
#' @noRd
fetch_kegg_mapping <- function(pathways, batch_size, on_error) {
  assert_that(is.character(pathways))
  batches <- split(pathways, ceiling(seq_along(pathways) / batch_size))

  raise_error <- FALSE
  pb <- progress::progress_bar$new(total = length(batches))
  tb <- purrr::map(batches, function(batch) {
    dbentries <- paste(batch, collapse = "+")

    # this returns a flat file
    resp <- http_request(get_kegg_url(), stringr::str_glue("get/{dbentries}"))
    if(resp$is_error) {
      catch_error("KEGG", resp, on_error)
      raise_error <<- TRUE
      return(NULL)
    }

    pb$tick()
    st <- httr2::resp_body_string(resp$response)
    parse_kegg_genes(st)
  }) |>
    purrr::list_rbind()

  if(raise_error) {
    return(NULL)
  } else {
    return(tb)
  }
}


#' Get functional term data from KEGG
#'
#' Download information (pathway ID and name) and gene-pathway mapping (entrez
#' gene ID, gene symbol and pathway ID) from KEGG. Gene symbols are extracted
#' from gene descriptions. For some species (e.g. yeast), gene symbols are
#' returned instead of entrez IDs and not in gene description.
#'
#' @param species KEGG species code, for example "hsa" for human. The
#'   full list of available KEGG species can be found by using \code{fetch_kegg_species}.
#' @param batch_size Number of pathways sent to KEGG database in one query. The
#'   maximum allowed is 10.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A list with \code{terms} and \code{mapping} tibbles.
#' @importFrom assertthat assert_that is.count
#' @export
#' @examples
#' kegg_data <- fetch_kegg("mge", on_error = "warn")
fetch_kegg <- function(species, batch_size = 10, on_error = c("stop", "warn", "ignore")) {
  on_error <- match.arg(on_error)

  assert_that(!missing(species), msg = "Argument 'species' is missing.")
  assert_that(is.count(batch_size))
  assert_that(batch_size <= 10, msg = "batch_size needs to be between 1 and 10")

  terms <- fetch_kegg_pathways(species, on_error)
  if(is.null(terms))
    return(NULL)
  mapping <- fetch_kegg_mapping(terms$term_id, batch_size, on_error)

  list(
    terms = terms,
    mapping = mapping
  )
}

test_that("Incorrect species in fetch_kegg", {
  expect_error(fetch_kegg())
  expect_error(fetch_kegg(1243))
  expect_error(fetch_kegg("not a species"))
})

test_that("Incorrect batch size in fetch_kegg", {
  expect_error(fetch_kegg("sce", "aha!"))
  expect_error(fetch_kegg("sce", -1))
  expect_error(fetch_kegg("sce", 0))
  expect_error(fetch_kegg("sce", 1000))
})


test_that("Parsing KEGG flat file", {
  flat <- readr::read_file("../test_data/kegg_test.txt")
  expected <- readr::read_rds("../test_data/kegg_parsing_result.rds")
  parsed <- parse_kegg_genes(flat)
  expect_equal(parsed, expected)
})


test_that("Expected behaviour from a non-responsive server", {
  species <- "mge"
  pathway <- "mge00010"
  httr2::with_mocked_responses(
    mock = mocked_500,
    code = {
      test_unresponsive_server(fetch_kegg_species)
      test_unresponsive_server(fetch_kegg_pathways, species = species)
      test_unresponsive_server(fetch_kegg_mapping, pathways = pathway, batch_size = 1)
      test_unresponsive_server(fetch_kegg, species = species)
    })
})


test_that("Expected return from fetch_kegg_species", {
  expected_selection <- c("hsa", "mmu", "rno", "sce", "dme", "cel")
  spec <- fetch_kegg_species(on_error = "ignore")
  if(!is.null(spec)) {
    expect_is(spec, "tbl")
    expect_true(all(expected_selection %in% spec$designation))
  }
})


test_that("Correct response from fetch_kegg", {
  species <- "mge"

  expected_terms <- tibble::tribble(
    ~term_id, ~term_name,
    "mge01100", "Metabolic pathways - Mycoplasmoides genitalium G37",
    "mge01230", "Biosynthesis of amino acids - Mycoplasmoides genitalium G37",
    "mge03010", "Ribosome - Mycoplasmoides genitalium G37",
    "mge03030", "DNA replication - Mycoplasmoides genitalium G37"
  )

  expected_mapping <- tibble::tribble(
    ~term_id, ~gene_symbol,
    "mge00010", "pgi",
    "mge00010", "fba",
    "mge03010", "rpsB",
    "mge03030", "polC"
  )

  re <- fetch_kegg(species, on_error = "ignore")
  if(!is.null(re)) {
    test_fetched_structure(re)
    test_terms(re$terms, expected_terms)
    test_mapping(re$mapping, expected_mapping, "gene_symbol")
  }
})

#' Example set of background genes.
#'
#' A set of gene names for proteins that could be detected in a typical
#' proteomics experiment on yeast samples.
#'
#' @usage data(exmpl_all)
#' @format A character vector with 6985 elements.
#' @return A vector with background gene names.
"exmpl_all"


#' Example set of selected genes.
#'
#' A set of gene names manually selected to illustrate functional enrichment.
#'
#' @usage data(exmpl_sel)
#' @format A character vector with 21 elements.
#' @return A vector with selected gene names.
"exmpl_sel"


#' Differential expression results for yeast RNA-seq.
#'
#' A subset of 6 + 6 replicates was selected from data set reported in
#' https://doi.org/10.1093/bioinformatics/btv425
#'
#' @usage data(yeast_de)
#' @format A tibble with 5 columns
#' @return Results for differential expression for yeast RNA-seq.
"yeast_de"


#' GO-terms data downloaded for the vignette.
#'
#' Downloaded using \code{go <- fetch_go(species = "sgd")}
#'
#' @usage data(go)
#' @format A list of two tibbles
#' @return Contains GO-term descriptions and gene mapping.
"go"


#' GO species
#'
#' Downloaded using \code{go_species <- fetch_go_species()}
#'
#' @usage data(go_species)
#' @format A tibble
#' @return Contains species available through Gene Ontology
"go_species"

#' URL of GO obo top dir
#'
#' @return A string with URL.
#' @noRd
get_go_obo_dir <- function() {
  getOption("GO_OBO_DIR", "http://purl.obolibrary.org/obo")
}

#' URL of GO gene association file
#'
#' @return A string with URL.
#' @noRd
get_go_obo_file <- function() {
  getOption("GO_OBO_FILE", "http://purl.obolibrary.org/obo/go.obo")
}

#' URL of GO species webpage
#'
#' @return A string with URL.
#' @noRd
get_go_species_url <- function() {
  getOption("GO_SPECIES_URL", "http://current.geneontology.org/products/pages/downloads.html")
}

#' URL of GO annotation server
#'
#' @return A string with URL.
#' @noRd
get_go_annotation_url <- function() {
  getOption("GO_ANNOTATION_URL", "http://current.geneontology.org/annotations")
}

#' URL of Ensembl's biomart
#'
#' @return A string with URL.
#' @noRd
get_biomart_url <- function() {
  getOption("ENSEMBL_BIOMART", "http://www.ensembl.org")
}

# See https://www.ensembl.org/info/data/biomart/biomart_restful.html for more details.
get_biomart_xml <- function(dataset) {
stringr::str_glue("
<?xml version='1.0' encoding='UTF-8'?>
<!DOCTYPE Query>
<Query virtualSchemaName = 'default' uniqueRows = '1' count='0' datasetConfigVersion='0.6' header='1' formatter='TSV' requestid='biomaRt'>
  <Dataset name = '{dataset}'>
    <Attribute name = 'ensembl_gene_id'/>
    <Attribute name = 'external_gene_name'/>
    <Attribute name = 'go_id'/>
    <Attribute name = 'go_linkage_type'/>
  </Dataset>
</Query>") |>
    stringr::str_replace_all("\n", "") |>
    stringr::str_replace_all("\\>\\s+\\<", "><")
}

#' Parse OBO file and return a tibble with key and value
#'
#' @param obo Obo file content as a character vector
#'
#' @return A tibble with term_id, key and value
#' @noRd
parse_obo_file <- function(obo) {
  # Find index of start and end line of each term

  # Start lines
  starts <- stringr::str_which(obo, "\\[Term\\]")

  # Empty lines at end of each term
  blanks <- stringr::str_which(obo, "^$")
  blanks <- blanks[blanks > starts[1]]

  # No space at the end
  if(length(blanks) < length(starts))
    blanks <- c(blanks, length(obo) + 1)

  # End lines: ignore empty lines beyond terms
  ends <- blanks[seq_len(length(starts))]

  # Parse each term
  purrr::map2_dfr(starts, ends, function(i1, i2) {
    obo_term <- obo[(i1 + 1):(i2 - 1)]
    trm <- obo_term |>
      stringr::str_split(":\\s", 2, simplify = TRUE)
    colnames(trm) <- c("key", "value")
    # assuming term_id is in the first line, if not, we are screwed
    tid <- trm[1, 2]
    cbind(trm, term_id = tid) |>
      as.data.frame()
  }) |>
    tibble::as_tibble()
}

#' Extract GO term IDs and names from parsed OBO data
#'
#' @param parsed OBO data parsed by \code{parse_obo_file}
#'
#' @return A tibble with term_id and term_name
#' @noRd
extract_obo_terms <- function(parsed) {
  # Binding variables from non-standard evaluation locally
  key <- term_id <- value <- name <- term_name <- namespace <- NULL

  terms <- parsed |>
    dplyr::filter(key %in% c("name", "namespace")) |>
    tidyr::pivot_wider(id_cols = term_id, names_from = key, values_from = value) |>
    dplyr::rename(term_name = name)

  alt_terms <- parsed |>
    dplyr::filter(key == "alt_id") |>
    dplyr::left_join(terms, by = "term_id") |>
    dplyr::select(term_id = value, term_name, namespace)

  dplyr::bind_rows(
    terms,
    alt_terms
  )
}


#' Download GO term descriptions
#'
#' @param use_cache Logical, if TRUE, the remote file will be cached locally.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A tibble with term_id and term_name.
#' @noRd
fetch_go_terms <- function(use_cache, on_error) {
  if(!assert_url_path(get_go_obo_dir(), on_error))
    return(NULL)

  obo_file <- get_go_obo_file()
  lpath <- cached_url_path("obo", obo_file, use_cache)
  readr::read_lines(lpath) |>
    parse_obo_file() |>
    extract_obo_terms()
}


#' Find all species available from geneontology.org
#'
#' This function returns a fixed table of Gene Ontology species designations
#' included with the package. It is a temporary compatibility patch for legacy
#' Gene Ontology GAF file names such as \file{goa_human.gaf.gz},
#' \file{mgi.gaf.gz}, and \file{sgd.gaf.gz}. These legacy file names are still
#' served by Gene Ontology, but they are no longer listed on the annotation
#' downloads page that this function previously scraped.
#'
#' The returned table may therefore become stale if Gene Ontology removes or
#' changes the legacy download paths. A future version of \pkg{fenr} may replace
#' this snapshot with a different discovery mechanism or with the newer
#' Gene Ontology annotation file names.
#'
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'   This argument is retained for API compatibility; the current implementation
#'   reads a packaged data table rather than querying a remote server.
#'
#' @return A tibble with columns \code{species} and \code{designation}.
#' @export
#' @examples
#' go_species <- fetch_go_species(on_error = "warn")
fetch_go_species <- function(on_error = c("stop", "warn", "ignore")) {
  on_error <- match.arg(on_error)

  go_species <- NULL

  utils::data("go_species", package = "fenr", envir = environment())
  return(go_species)
}


#' Download GO term gene mapping from geneontology.org
#'
#' @param species Species designation. Base file name for species file under
#'   \url{http://current.geneontology.org/annotations}. Examples are
#'   \file{goa_human} for human, \file{mgi} for mouse or \file{sgd} for yeast.
#' @param use_cache Logical, if TRUE, the remote file will be cached locally.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A tibble with columns \code{gene_symbol}, \code{uniprot_id}, 
#'   \code{term_id} and \code{evidence}.
#' @noRd
fetch_go_genes_go <- function(species, use_cache, on_error) {
  # Binding variables from non-standard evaluation locally
  gene_id <- db_object_synonym <- symbol <- NULL
  db_id <- go_term <- evidence <- NULL

  url <- get_go_annotation_url()
  gaf_file <- stringr::str_glue("{url}/{species}.gaf.gz")
  if(!assert_url_path(gaf_file, on_error))
    return(NULL)

  lpath <- cached_url_path(stringr::str_glue("go_gaf_{species}"), gaf_file, use_cache)
  readr::read_tsv(lpath, comment = "!", quote = "", col_names = GAF_COLUMNS,
                  col_types = GAF_TYPES) |>
    dplyr::mutate(gene_id = stringr::str_remove(db_object_synonym, "\\|.*$")) |>
    dplyr::select(gene_symbol = symbol, gene_id, db_id, term_id = go_term, evidence) |>
    dplyr::distinct()
}


#' Get functional term data from gene ontology
#'
#' Download term information (GO term ID and name) and gene-term mapping (gene
#' symbol and GO term ID) from gene ontology.
#'
#' @details This function relies on Gene Ontology's GAF files containing more
#'   generic information than gene symbols. Here, the third column of the GAF
#'   file (DB Object Symbol) is returned as \code{gene_symbol}, but, depending
#'   on the \code{species} argument it can contain other entities, e.g. RNA or
#'   protein complex names. Similarly, the eleventh column of the GAF file (DB
#'   Object Synonym) is returned as \code{gene_id}. It is up to the user to
#'   select the appropriate database.
#'
#' @param species Species designation. Examples are \file{goa_human} for human,
#'   \file{mgi} for mouse or \file{sgd} for yeast. Full list of available
#'   species can be obtained using \code{fetch_go_species} - column
#'   \code{designation}.
#' @param use_cache Logical, if TRUE, the remote file will be cached locally.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A list with \code{terms} and \code{mapping} tibbles.
#' @importFrom assertthat assert_that
#' @noRd
fetch_go_from_go <- function(species, use_cache, on_error) {
  assert_that(!missing(species), msg = "Argument 'species' is missing.")
  assert_species(species, "fetch_go_species", on_error)

  mapping <- fetch_go_genes_go(species = species, use_cache = use_cache, on_error = on_error)
  if(is.null(mapping))
    return(NULL)

  terms <- fetch_go_terms(use_cache = use_cache, on_error = on_error)
  if(is.null(terms))
    return(NULL)

  list(
    terms = terms,
    mapping = mapping
  )
}



#' Download GO term gene mapping from Ensembl
#'
#' @param dataset Dataset you want to use. To see the different datasets
#'   available within a biomaRt you can e.g. do: mart <-
#'   biomaRt::useEnsembl(biomart = "ensembl"), followed by
#'   biomaRt::listDatasets(mart).
#' @param use_cache Logical, if TRUE, the remote data will be cached locally.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A tibble with columns \code{gene_id}, \code{gene_symbol},
#'   \code{term_id} and \code{evidence}.
#' @noRd
fetch_go_genes_bm <- function(dataset, use_cache, on_error) {
  xml <- get_biomart_xml(dataset) |>
    stringr::str_replace_all("\\s", "%20")
  biomart_path <- paste0(get_biomart_url(), "/biomart/martservice")
  if(!assert_url_path(biomart_path, on_error))
    return(NULL)

  req <- paste0(biomart_path, "?query=", xml)

  # Problems with cache, bfcneedsupdate returns error for this query
  # lpath <- cached_url_path(stringr::str_glue("biomart_{dataset}"), resp, use_cache)
  res <- readr::read_tsv(req, show_col_types = FALSE)
  if(ncol(res) == 4) {
    res |> rlang::set_names(c("gene_id", "gene_symbol", "term_id", "evidence"))
  } else {
    error_response("Problem with Biomart", on_error)
  }
}



#' Get functional term data from Ensembl
#'
#' Download term information (GO term ID and name) and gene-term mapping
#' (gene ID, symbol and GO term ID) from Ensembl.
#'
#' @param dataset Dataset you want to use. To see the different datasets
#'   available within a biomaRt you can e.g. do: mart <-
#'   biomaRt::useEnsembl(biomart = "ensembl"), followed by
#'   biomaRt::listDatasets(mart).
#' @param use_cache Logical, if TRUE, the remote file will be cached locally.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @importFrom assertthat assert_that is.string
#' @return A list with \code{terms} and \code{mapping} tibbles.
#' @noRd
fetch_go_from_bm <- function(dataset, use_cache, on_error) {
  assert_that(!missing(dataset), msg = "Argument 'dataset' is missing.")
  assert_that(is.string(dataset))

  mapping <- fetch_go_genes_bm(dataset, use_cache = use_cache, on_error = on_error)
  if(is.null(mapping))
    return(error_response("Could not retrieve mapping from Ensembl", on_error))
  terms <- fetch_go_terms(use_cache = use_cache, on_error)

  list(
    terms = terms,
    mapping = mapping
  )
}


#' Get Gene Ontology (GO) data
#'
#'
#' This function downloads term information (GO term ID and name) and gene-term
#' mapping (gene ID, symbol, and GO term ID) from either the Ensembl database
#' (using BioMart) or the Gene Ontology database (using GAF files), depending on
#' the provided argument.
#'
#' @details If \code{species} is provided, mapping from a Gene Ontology GAF file
#'   will be downloaded. GAF files contain more generic information than gene
#'   symbols. In this function, the third column of the GAF file (DB Object
#'   Symbol) is returned as \code{gene_symbol}, but, depending on the
#'   \code{species} argument it can contain other entities, e.g. RNA or protein
#'   complex names. Similarly, the eleventh column of the GAF file (DB Object
#'   Synonym) is returned as \code{gene_id}. It is up to the user to select
#'   the appropriate database.
#'
#'   Alternatively, if \code{dataset} is provided, mapping will be downloaded
#'   from Ensembl database. It will gene symbol and Ensembl gene ID.
#'
#' @param species (Optional) Species designation. Examples are \code{goa_human}
#'   for human, \code{mgi} for mouse, or \code{sgd} for yeast. Full list of
#'   available species can be obtained using \code{fetch_go_species} - column
#'   \code{designation}. This argument is used when fetching data from the Gene
#'   Ontology database.
#' @param dataset (Optional) A string representing the dataset passed to
#'   Ensebml's Biomart, e.g. 'scerevisiae_gene_ensembl'. To see the different
#'   datasets available within a biomaRt you can e.g. do: mart <-
#'   biomaRt::useEnsembl(biomart = "ensembl"), followed by
#'   biomaRt::listDatasets(mart).
#' @param use_cache Logical, if TRUE, the remote data will be cached locally.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A list with \code{terms} and \code{mapping} tibbles.
#' @export
#' @importFrom assertthat assert_that
#' @examples
#' # Fetch GO data from Ensembl
#' go_data_ensembl <- fetch_go(dataset = "scerevisiae_gene_ensembl", on_error = "warn")
#' # Fetch GO data from Gene Ontology
#' go_data_go <- fetch_go(species = "sgd", on_error = "warn")
fetch_go <- function(species = NULL, dataset = NULL, use_cache = TRUE,
                     on_error = c("stop", "warn", "ignore")) {
  on_error <- match.arg(on_error)

  assert_that(!(is.null(species) & is.null(dataset)),
              msg = "One of the arguments 'species' or 'dataset' must be supplied.")
  assert_that(is.null(species) | is.null(dataset),
              msg = "Only one of the arguments 'species' or 'dataset' must be supplied.")

  if (!is.null(species)) {
    fetch_go_from_go(species, use_cache = use_cache, on_error = on_error)
  } else {
    fetch_go_from_bm(dataset, use_cache = use_cache, on_error = on_error)
  }
}

test_that("Incorrect aruments in fetch_go", {
  expect_error(fetch_go())
  expect_error(fetch_go(species = "sgd", dataset = "scerevisiae_gene_ensembl"))
})


test_that("Incorrect species in fetch_go", {
  expect_error(fetch_go(species = 1243))
  expect_error(fetch_go(species = "not a species"))
})

test_that("Incorrect dataset triggers error", {
  expect_error(fetch_go(dataset = "not a dataset"))
  expect_error(fetch_go_from_bm(dataset = "not a dataset", use_cache = FALSE, on_error = "stop"))
})

test_that("Processing OBO file", {
  expected_ids <- c("GO:0000001", "GO:0000022", "GO:0000232", "GO:1905121")
  expected_names <- c("mitochondrion inheritance", "mitotic spindle elongation",
                      "obsolete nuclear interphase chromosome", "mitotic spindle elongation")

  trms <- readr::read_lines("../test_data/go_obo_test.txt") |>
    parse_obo_file() |>
    extract_obo_terms()

  expect_s3_class(trms, "tbl")
  expect_equal(trms$term_id, expected_ids)
  expect_equal(trms$term_name, expected_names)
})


test_that("Expected behaviour from a non-responsive server", {
  species <- "sgd"
  httr2::with_mocked_responses(
    mock = mocked_500,
    code = {
      test_unresponsive_server(fetch_go_terms, use_cache = FALSE)
      #test_unresponsive_server(fetch_go_species)
      test_unresponsive_server(fetch_go_from_go, species = species, use_cache = FALSE)
      test_unresponsive_server(fetch_go_from_bm, dataset = "scerevisiae_gene_ensembl", use_cache = FALSE)
      test_unresponsive_server(fetch_go_genes_bm, dataset = "scerevisiae_gene_ensembl", use_cache = FALSE)
    })
})


test_that("Expected return from fetch_go_species", {
  expected_selection <- c("goa_human", "mgi", "rgd", "sgd", "fb", "wb")
  spec <- fetch_go_species(on_error = "ignore")
  if(!is.null(spec)) {
    expect_is(spec, "tbl")
    expect_true(all(expected_selection %in% spec$designation))
  }
})


test_that("GO yeast from GO is correct", {
  species <- "sgd"

  expected_terms <- tibble::tribble(
    ~term_id, ~term_name,
    "GO:0000166", "nucleotide binding",
    "GO:0006096", "glycolytic process",
    "GO:0005199", "structural constituent of cell wall",
    "GO:0004365", "glyceraldehyde-3-phosphate dehydrogenase (NAD+) (phosphorylating) activity"
  )

  expected_mapping <- tibble::tribble(
    ~term_id, ~gene_symbol,
    "GO:0000166", "POL1",
    "GO:0006096", "FBA1",
    "GO:0005199", "CWP2",
    "GO:0004365", "TDH3"
  )

  re <- fetch_go(species = species, on_error = "ignore")
  if(!is.null(re)) {
    test_fetched_structure(re)
    test_terms(re$terms, expected_terms)
    test_mapping(re$mapping, expected_mapping, "gene_symbol")
  }
})


test_that("GO yeast from Ensembl is correct", {
  dataset <- "scerevisiae_gene_ensembl"

  expected_terms <- tibble::tribble(
    ~term_id, ~term_name,
    "GO:0000166", "nucleotide binding",
    "GO:0006096", "glycolytic process",
    "GO:0005199", "structural constituent of cell wall",
    "GO:0004365", "glyceraldehyde-3-phosphate dehydrogenase (NAD+) (phosphorylating) activity"
  )

  expected_mapping <- tibble::tribble(
    ~term_id, ~gene_id,
    "GO:0000166", "YNL102W",
    "GO:0006096", "YKL060C",
    "GO:0005199", "YKL096W-A",
    "GO:0004365", "YGR192C"
  )

  re <- fetch_go(dataset = dataset, on_error = "ignore")

  if(!is.null(re)) {
    test_fetched_structure(re)
    test_terms(re$terms, expected_terms)
    test_mapping(re$mapping, expected_mapping, "gene_id")
  }
})

#' Reactome server base URL
#'
#' @return A string with URL.
#' @noRd
get_reactome_url <- function() {
  getOption("REACTOME_BASE_URL", "https://reactome.org/ContentService")
}

#' URL of Reactome download top dir
#'
#' @return A string with URL.
#' @noRd
get_reactome_download <- function() {
  getOption("REACTOME_DOWNLOAD", "https://reactome.org/download/current/")
}


#' URL of Reactome Ensembl file
#'
#' @return A string with URL.
#' @noRd
get_reactome_ensembl_file <- function() {
  getOption("REACTOME_ENSEMBL_FILE", "https://reactome.org/download/current/Ensembl2Reactome_PE_Pathway.txt")
}

#' URL of Reactome gene association file
#'
#' @return A string with URL.
#' @noRd
get_reactome_gaf_file <- function() {
  getOption("REACTOME_GAF_FILE", "https://reactome.org/download/current/gene_association.reactome.gz")
}

#' List of available Reactome species
#'
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A tibble with species names used by Reactome.
#' @export
#' @examples
#' re <- fetch_reactome_species(on_error = "warn")
fetch_reactome_species <- function(on_error = c("stop", "warn", "ignore")) {
  on_error <- match.arg(on_error)

  # Binding variables from non-standard evaluation locally
  dbId <- displayName <- taxId <- NULL

  resp <- http_request(get_reactome_url(), "data/species/main")
  if(resp$is_error)
    return(catch_error("Reactome", resp, on_error))

  httr2::resp_body_json(resp$response) |>
    purrr::map(tibble::as_tibble) |>
    purrr::list_rbind() |>
    dplyr::select(db_id = dbId, designation = displayName, tax_id = taxId) |>
    dplyr::distinct()
}


#' Download pathway data from Reactome
#'
#' @param tax_id Taxonomy ID of the species, a string
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A tibble with columns \code{term_id} and \code{term_name}
#' @noRd
fetch_reactome_pathways <- function(tax_id, on_error) {
  # Binding variables from non-standard evaluation locally
  stId <- displayName <- NULL

  path <- "data/schema/Pathway/min"
  params <- list(
    species = tax_id,
    page = 1,
    offset = 20000
  )

  resp <- http_request(get_reactome_url(), path, parameters = params)
  if(resp$is_error)
    return(catch_error("Reactome", resp, on_error))

  httr2::resp_body_json(resp$response) |>
    purrr::map(tibble::as_tibble) |>
    purrr::list_rbind() |>
    dplyr::select(term_id = stId, term_name = displayName)
}


#' Download term - Ensembl gene ID mapping from Reactome
#'
#' @details This function downloads one large file containing a mapping between
#'   Enxembl gene IDs and Reactome terms. This is significantly faster than
#'   \code{fetch_reactome_genes}, but if gene symbols are required, needs
#'   additional ID conversion.
#'
#' @param spec Reactome species.
#' @param use_cache Logical, if TRUE, the remote file will be cached locally.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A tibble with columns \code{gene_id} and \code{term_id}
#' @noRd
fetch_reactome_ensembl_genes <- function(spec, use_cache, on_error) {
  # Binding variables from non-standard evaluation locally
  species <- gene_id <- gene_name <- gene_symbol <- term_id <- NULL

  reactome_topdir <- get_reactome_download()
  if(!assert_url_path(reactome_topdir, on_error))
    return(NULL)

  ensembl_file <- get_reactome_ensembl_file()
  lpath <- cached_url_path("ensembl2reactome", ensembl_file, use_cache)
  colms <- c("gene_id", "reactome_gene_id", "gene_name", "term_id", "url", "event", "evidence", "species")
  readr::read_tsv(lpath, col_names = colms, show_col_types = FALSE) |>
    dplyr::filter(species == spec) |>
    dplyr::mutate(gene_symbol = stringr::str_remove(gene_name, "\\s.+$")) |>
    dplyr::select(gene_id, gene_symbol, term_id) |>
    # Apart from gene symbols there are other entities in Reactome
    dplyr::filter(stringr::str_detect(gene_symbol, "^[A-Z0-9]+$")) |>
    dplyr::distinct()
}

#' Download term - gene association from Reactome
#'
#' @param tax_id Taxon ID
#' @param use_cache Logical, if TRUE, the remote file will be cached locally.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A tibble with columns \code{accession_number}, \code{gene_symbol} and \code{term_id}
#' @noRd
fetch_reactome_gene_association <- function(tax_id, use_cache, on_error) {
  # Binding variables from non-standard evaluation locally
  symbol <- taxon <- db_ref <- db_id <- NULL

  if(!assert_url_path(get_reactome_download(), on_error))
    return(NULL)
  gaf_file <- get_reactome_gaf_file()

  lpath <- cached_url_path("reactome_gaf", gaf_file, use_cache)
  readr::read_tsv(lpath, comment = "!", quote = "", col_names = GAF_COLUMNS, col_types = GAF_TYPES, skip = 4) |>
    dplyr::mutate(
      symbol = stringr::str_remove(symbol, "_.+$"),
      taxon = stringr::str_remove(taxon, "taxon:"),
      db_ref = stringr::str_remove(db_ref, "REACTOME:")
    ) |>
    dplyr::filter(stringr::str_detect(db_ref, "^R-") & taxon == tax_id) |>
    dplyr::select(accession_number = db_id, gene_symbol = symbol, term_id = db_ref) |>
    dplyr::distinct()
}


#' Download term - gene symbol mapping from Reactome
#'
#' @details This function interrogates Reactome API to get term-gene mapping for
#'   all pathways. This is considerable slower than
#'   \code{fetch_reactome_ensembl_genes}. Warning, occasionally, for some
#'   pathways, Reactome does not return gene symbol - only UniProt accession
#'   number is available.
#'
#' @param pathways A character vector with Reactome patway IDs to get
#'   corresponding genes from.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A tibble with columns\code{term_id}, \code{accession_number} and
#'   \code{gene_symbol}.
#' @noRd
fetch_reactome_api_genes <- function(pathways, on_error) {
  # Binding variables from non-standard evaluation locally
  database_name <- gene_symbol <- NULL

  raise_error <- FALSE
  pb <- progress::progress_bar$new(total = length(pathways))
  tb <- purrr::map(pathways, function(pathway) {
    pb$tick()
    path <- stringr::str_glue("data/participants/{pathway}/referenceEntities")
    resp <- http_request(get_reactome_url(), path)
    if(resp$is_error) {
      raise_error <<- TRUE
      return(catch_error("Reactome", resp, on_error))
    }

    httr2::resp_body_json(resp$response) |>
      purrr::map(function(js) {
        tibble::tibble(
          database_name = js$databaseName,
          gene_symbol = js$geneName,
          accession_number = js$identifier
        )
      }) |>
      purrr::list_rbind() |>
      dplyr::filter(database_name == "UniProt") |>
      dplyr::select(-database_name) |>
      tidyr::unnest(gene_symbol) |>
      tibble::add_column(term_id = pathway, .before = 1)
  }) |>
    purrr::list_rbind()
  if(raise_error) {
    return(NULL)
  } else {
    return(tb)
  }
}


#' Get functional term data from Reactome
#'
#' Download term information (pathway ID and name) and gene-pathway mapping
#' (Ensembl gene ID or gene symbol and pathway ID) from Reactome.
#'
#' @details Reactome makes mapping between Ensembl ID and pathway ID available
#'   in form of one downloadable file. This mapping contains gene symbols as
#'   well. Also, a gene association file with mapping between UniProt accession
#'   number, gene symbol and Reactome term is available. If \code{source =
#'   "ensembl"} or \code{source =
#'   "gene_association"} is set, one large file will be downloaded and parsed.
#'   If \code{source = "api"} is set, then Reactome APIs will be interrogated
#'   for each pathway available. This method is considerably slower, especially
#'   for large genomes. However, gene association file contains far fewer
#'   mappings than can be extracted using API. If gene symbols are needed, we
#'   recommend using  \code{source = "ensembl"}.
#'
#' @param species Reactome species designation, for example "Homo sapiens" for
#'   human. Full list of available species can be found using
#'   \code{fetch_reactome_species()}.
#' @param source How to download the mapping. If 'ensembl' or
#'   'gene_association', one mapping file provided by Reactome will be
#'   downloaded, if 'api', then Reactome API will be used. See details.
#' @param use_cache Logical, if TRUE, the remote file will be cached locally.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#'
#' @return A list with \code{terms} and \code{mapping} tibbles
#' @importFrom assertthat assert_that
#' @export
#' @examples
#' reactome_data <- fetch_reactome("Saccharomyces cerevisiae", on_error = "warn")
fetch_reactome <- function(species, source = c("ensembl", "api", "gene_association"),
                           use_cache = TRUE, on_error = c("stop", "warn", "ignore")) {
  source <- match.arg(source)
  on_error <- match.arg(on_error)
  assert_that(!missing(species), msg = "Argument 'species' is missing.")

  tax_id <- match_species(species, "fetch_reactome_species", "tax_id", on_error)
  if(is.null(tax_id))
    return(NULL)

  terms <- fetch_reactome_pathways(tax_id, on_error)

  if (source == "ensembl") {
    mapping <- fetch_reactome_ensembl_genes(spec = species, use_cache = use_cache, on_error = on_error)
  } else if (source == "gene_association") {
    mapping <- fetch_reactome_gene_association(tax_id = tax_id, use_cache = use_cache, on_error = on_error)
  } else {
    mapping <- fetch_reactome_api_genes(terms$term_id, on_error = on_error)
  }

  list(
    terms = terms,
    mapping = mapping
  )
}

species <- "Saccharomyces cerevisiae"
tax_id <- "4932"

expected_pathways <- tibble::tribble(
  ~term_id, ~term_name,
  "R-SCE-6798695", "Neutrophil degranulation",
  "R-SCE-983168", "Antigen processing: Ubiquitination & Proteasome degradation",
  "R-SCE-204005", "COPII-mediated vesicle transport",
  "R-SCE-5668541", "TNFR2 non-canonical NF-kB pathway"
)

expected_ensembl <- tibble::tribble(
  ~term_id, ~gene_id,
  "R-SCE-70171", "YJL052W",
  "R-SCE-70171", "YGR192C",
  "R-SCE-68952", "YNL102W",
  "R-SCE-983168", "YKL022C"
)

expected_gene_association <- tibble::tribble(
  ~term_id, ~gene_symbol,
  "R-SCE-6793739", "FBRL",
  "R-SCE-9749345", "CDT1",
  "R-SCE-1252249", "ATPB",
  "R-SCE-9749381", "CDC6",
)

##################################################

test_that("Incorrect species in fetch_reactome", {
  expect_error(fetch_reactome())
  expect_error(fetch_reactome(1243))
  expect_error(fetch_reactome("not a species"))
  expect_error(fetch_reactome("Homo sapiens", "blah"))
})

test_that("Expected behaviour from a non-responsive server", {
  httr2::with_mocked_responses(
    mock = mocked_500,
    code = {
      test_unresponsive_server(fetch_reactome_species)
      test_unresponsive_server(fetch_reactome_pathways, tax_id = tax_id)
      test_unresponsive_server(fetch_reactome_api_genes, pathways = "R-SCE-68952")
      test_unresponsive_server(fetch_reactome, species = species)
      test_unresponsive_server(fetch_reactome_ensembl_genes, spec = species, use_cache = FALSE)
      test_unresponsive_server(fetch_reactome_gene_association, tax_id = tax_id, use_cache = FALSE)
    })
})

test_that("Expected return from fetch_reactome_species", {
  expected_selection <- c("Homo sapiens", "Mus musculus", "Rattus norvegicus",
                          "Saccharomyces cerevisiae", "Drosophila melanogaster", "Caenorhabditis elegans")
  spec <- fetch_reactome_species(on_error = "ignore")
  if(!is.null(spec)) {
    expect_is(spec, "tbl")
    expect_true(all(expected_selection %in% spec$designation))
  }
})


test_that("Correct Reactome Ensembl from yeast", {
  re <- fetch_reactome(species, source = "ensembl", use_cache = FALSE, on_error = "ignore")
  if(!is.null(re)) {
    expect_is(re, "list")
    expect_length(re, 2)
    expect_named(re)
    expect_equal(names(re), c("terms", "mapping"))

    # Check pathways
    paths <- re$terms
    expect_is(paths, "tbl")

    merged <- expected_pathways |>
      dplyr::left_join(paths, by = c("term_id", "term_name")) |>
      tidyr::drop_na()
    expect_equal(nrow(expected_pathways), nrow(merged))

    # Check mapping
    mapping <- re$mapping |>
      dplyr::select(gene_id, term_id) |>
      dplyr::distinct()
    expect_is(mapping, "tbl")
    merged <- expected_ensembl |>
      dplyr::left_join(mapping, by = c("term_id", "gene_id")) |>
      tidyr::drop_na()
    expect_equal(nrow(expected_ensembl), nrow(merged))
  }
})


test_that("Correct Reactome gene association from yeast", {
  re <- fetch_reactome(species, source = "gene_association", use_cache = FALSE, on_error = "ignore")
  if(!is.null(re)) {
    expect_is(re, "list")
    expect_length(re, 2)
    expect_named(re)
    expect_equal(names(re), c("terms", "mapping"))

    # Check mapping
    mapping <- re$mapping
    expect_is(mapping, "tbl")
    merged <- expected_gene_association |>
      dplyr::left_join(mapping, by = c("term_id", "gene_symbol")) |>
      tidyr::drop_na()
    expect_equal(nrow(expected_gene_association), nrow(merged))
  }
})

# API mapping is slow, so we use much smaller organism for tests

small_species <- "Mycobacterium tuberculosis"
expected_api <- tibble::tribble(
  ~term_id, ~gene_symbol,
  "R-MTU-964903", "aroA",
  "R-MTU-870392", "cysU",
  "R-MTU-870392", "sir",
  "R-MTU-879299", "mca"
)

test_that("Correct Reactome API from M. tuberculosis", {
  re <- fetch_reactome(species = small_species, source = "api", on_error = "ignore")
  mapping <- re$mapping
  if(!is.null(mapping)) {
    expect_is(mapping, "tbl")
    merged <- expected_api |>
      dplyr::left_join(mapping, by = c("term_id", "gene_symbol")) |>
      tidyr::drop_na()
    expect_equal(nrow(expected_api), nrow(merged))
  }
})

#' URL of Bioplanet top dir
#'
#' @return A string with URL.
#' @noRd
get_bioplanet_download <- function() {
  getOption("BIOPLANET_PATHWAY_DOWNLOAD", "https://tripod.nih.gov/bioplanet/download")
}

#' URL of Bioplanet pathway file
#'
#' @return A string with URL.
#' @noRd
get_bioplanet_pathway_file <- function() {
  getOption("BIOPLANET_PATHWAY_FILE", "https://tripod.nih.gov/bioplanet/download/pathway.csv")
}


#' Get functional term data from BioPlanet
#'
#' Download term information (term ID and name) and gene-pathway mapping
#' (NCBI gene ID, gene symbol and pathway ID) from BioPlanet.
#'
#' @param use_cache Logical, if TRUE, the remote file will be cached locally.
#' @param on_error A character string indicating the error handling strategy:
#'   either "stop" to halt execution, "warn" to issue a warning and return
#'   `NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
#' @return A list with \code{terms} and \code{mapping} tibbles.
#' @export
#' @examples
#' bioplanet_data <- fetch_bp(on_error = "warn")
fetch_bp <- function(use_cache = TRUE, on_error = c("stop", "warn", "ignore")) {
  on_error <- match.arg(on_error)

  # Binding variables from non-standard evaluation locally
  PATHWAY_ID <- PATHWAY_NAME <- GENE_ID <- GENE_SYMBOL <- NULL

  if(!assert_url_path(get_bioplanet_download(), on_error))
    return(NULL)

  pathway_file <- get_bioplanet_pathway_file()

  lpath <- cached_url_path("bioplanet_pathway", pathway_file, use_cache)
  paths <- readr::read_csv(lpath, show_col_types = FALSE)

  terms <- paths |>
    dplyr::select(term_id = PATHWAY_ID, term_name = PATHWAY_NAME) |>
    dplyr::distinct()

  mapping <- paths |>
    dplyr::select(term_id = PATHWAY_ID, ncbi_id = GENE_ID, gene_symbol = GENE_SYMBOL) |>
    dplyr::distinct()

  list(
    terms = terms,
    mapping = mapping
  )
}

expected_mapping <- tibble::tribble(
  ~term_id, ~gene_symbol,
  "bioplanet_1025", "CDK1",
  "bioplanet_120", "IL1A",
  "bioplanet_1755", "RPL6",
  "bioplanet_1121", "POLR1A"
)

test_that("Bioplanet mapping makes sense", {
  bp <- fetch_bp(on_error = "ignore")
  if(!is.null(bp)) {
    expect_is(bp, "list")
    expect_setequal(names(bp), c("terms", "mapping"))
    mapping <- bp$mapping

    merged <- expected_mapping |>
      dplyr::left_join(mapping, by = c("term_id", "gene_symbol")) |>
      tidyr::drop_na()

    expect_equal(nrow(expected_mapping), nrow(merged))
  }
})


test_that("Expected behaviour from a non-responsive server", {
  httr2::with_mocked_responses(
    mock = mocked_500,
    code = {
      test_unresponsive_server(fetch_bp, use_cache = FALSE)
    })
})

## ----setup, include = FALSE-------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
options(width = 72)

## ----bioconductor_install, eval=FALSE---------------------------------
# if (!require("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install("fenr")

## ----figure, fig.cap="Overview of fenr data pipeline.", echo=FALSE----
knitr::include_graphics("figures/overview.png")

## ----load_fenr--------------------------------------------------------
library(fenr)
data(exmpl_all, exmpl_sel)

## ----load_go_data, include=FALSE--------------------------------------
library(tibble)
data(go)
data(go_species)

## ----fetch_go, eval=FALSE---------------------------------------------
# go <- fetch_go(species = "sgd")

## ----species, eval=FALSE----------------------------------------------
# go_species <- fetch_go_species()

## ----show_species-----------------------------------------------------
go_species

## ----go_terms---------------------------------------------------------
go$terms

## ----go_mapping-------------------------------------------------------
go$mapping

## ----prepare_for_enrichment-------------------------------------------
go_terms <- prepare_for_enrichment(go$terms, go$mapping, exmpl_all,
                                   feature_name = "gene_symbol")

## ----enrichment-------------------------------------------------------
enr <- functional_enrichment(exmpl_all, exmpl_sel, go_terms)

## ----enrichment_result------------------------------------------------
enr |>
  head(10)

## ----interactive_prepare, eval=FALSE----------------------------------
# data(yeast_de)
# term_data <- fetch_terms_for_example(yeast_de)

## ----shiny_app, eval=FALSE--------------------------------------------
# enrichment_interactive(yeast_de, term_data)

## ----topGO------------------------------------------------------------
suppressPackageStartupMessages({
  library(topGO)
})

# Preparing the gene set
all_genes <- setNames(rep(0, length(exmpl_all)), exmpl_all)
all_genes[exmpl_all %in% exmpl_sel] <- 1

# Define the gene selection function
gene_selection <- function(genes) {
  genes > 0
}

# Mapping genes to GO, we use the go2gene downloaded above and convert it to a
# list
go2gene <- go$mapping |> 
  dplyr::select(gene_symbol, term_id) |> 
  dplyr::distinct() |> 
  dplyr::group_by(term_id) |> 
  dplyr::summarise(ids = list(gene_symbol)) |> 
  tibble::deframe()

# Setting up the GO data for analysis
go_data <- new(
  "topGOdata",
  ontology = "BP",
  allGenes = all_genes,
  geneSel = gene_selection,
  annot = annFUN.GO2genes,
  GO2genes = go2gene
)

# Performing the enrichment test
fisher_test <- runTest(go_data, algorithm = "classic", statistic = "fisher")

## ----fenr_speed, eval = FALSE-----------------------------------------
# # Setting up data for enrichment
# go <- fetch_go(species = "sgd")
# go_terms <- prepare_for_enrichment(go$terms, go$mapping, exmpl_all,
#                                    feature_name = "gene_symbol")
# 
# # Executing the enrichment test
# enr <- functional_enrichment(exmpl_all, exmpl_sel, go_terms)

## ----session_info, echo=FALSE-----------------------------------------
sessionInfo()

timeout_url <- "http://www.google.com:81"


test_that("Correct reaction to a time-out in http_resuest()", {
  resp <- http_request(timeout_url, "", timeout = 1)
  expect_is(resp, "list")
  expect_equal(resp$is_error, TRUE)
  expect_equal(resp$status, "unknown")
})

test_that("Correct reaction to a time-out in assert_url_path()", {
  expect_warning({assert_url_path(timeout_url, on_error = "warn", timeout = 1)})
  resp <- assert_url_path(timeout_url, on_error = "ignore", timeout = 1)
  expect_equal(resp, FALSE)
})

---
title: "Fast functional enrichment"
author: "Marek Gierlinski"
package: fenr
output:
  BiocStyle::html_document:
    toc_float: true
abstract: |
  The *fenr* R package enables rapid functional enrichment analysis, typically completing in a fraction of a second, which makes it well-suited for interactive applications, such as Shiny apps. To accomplish this, fenr pre-downloads functional data (e.g., GO terms or KEGG pathways) and stores them in a format optimized for swift analysis of any arbitrary selection of features, including genes or proteins.
vignette: >
  %\VignetteIndexEntry{Fast functional enrichment}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
options(width = 72)
```


# Purpose

Functional enrichment analysis determines if specific biological functions or pathways are overrepresented in a set of features (e.g., genes, proteins). These sets often originate from differential expression analysis, while the functions and pathways are derived from databases such as *GO*, *Reactome*, or *KEGG*. In its simplest form, enrichment analysis employs Fisher's test to evaluate if a given function is enriched in the selection. The null hypothesis asserts that the proportion of features annotated with that function is the same between selected and non-selected features.

Functional enrichment analysis requires downloading large datasets from the aforementioned databases before conducting the actual analysis. While downloading data is time-consuming, Fisher's test can be performed rapidly. This package aims to separate these two steps, enabling fast enrichment analysis for various feature selections using a given database. It is specifically designed for interactive applications like Shiny. A small Shiny app, included in the package, demonstrates the usage of *fenr*.

## Caveats

Functional enrichment analysis should not be considered the ultimate answer in understanding biological systems. In many instances, it may not provide clear insights into biology. Specifically, when arbitrary groups of genes are selected, enrichment analysis only reveals the statistical overrepresentation of a functional term within the selection, which may not directly correspond to biological relevance. This package serves as a tool for data exploration; any conclusions drawn about biology require independent validation and further investigation.

# Installation

*fenr* can be installed from *Bioconductor* by using:

```{r bioconductor_install, eval=FALSE}
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("fenr")
```

# Overview

The input data for *fenr* consists of two tibbles or data frames (see figure below). One tibble contains functional term descriptions, while the other provides the term-to-feature mapping. Users can acquire these datasets using the convenient helper functions like `fetch_*`, which facilitate data retrieval from GO, KEGG, BioPlanet, or WikiPathways. Alternatively, users have the option to provide their own datasets. These tibbles are subsequently transformed into a fenr_terms object, specifically designed for efficient data retrieval and access. The resulting object is then employed by the functional_enrichment function to perform enrichment analysis for any chosen feature selection.

```{r figure, fig.cap="Overview of fenr data pipeline.", echo=FALSE}
knitr::include_graphics("figures/overview.png")
```

# Example

Package *fenr* and example data are loaded with the following commands:

```{r load_fenr}
library(fenr)
data(exmpl_all, exmpl_sel)
```

## Data preparation

The initial step involves downloading functional term data. *fenr* supports data downloads from *Gene Ontology*, *Reactome*, *KEGG*, *BioPlanet*, and *WikiPathways*. Custom ontologies can also be used, provided they are converted into an appropriate format (refer to the `prepare_for_enrichment` function for more information). The command below downloads functional terms and gene mapping from Gene Ontology (GO):

```{r load_go_data, include=FALSE}
library(tibble)
data(go)
data(go_species)
```

```{r fetch_go, eval=FALSE}
go <- fetch_go(species = "sgd")
```

`sgd` is a designation for yeast in Gene Ontology. The full list of available species designations can be obtain using the command:

```{r species, eval=FALSE}
go_species <- fetch_go_species()
```

which returns a tibble:

```{r show_species}
go_species
```

`go` is a list with two tibbles. The first tibble contains term information:

```{r go_terms}
go$terms
```

The second tibble contains gene-term mapping:

```{r go_mapping}
go$mapping
```

Note that the mapping can be filtered based on the [evidence code](https://geneontology.org/docs/guide-go-evidence-codes/) (column `evidence`) to include only high-quality GO annotations, before further analysis. Here, we simply use all annotations.

To make these user-friendly data more suitable for rapid functional enrichment analysis, they need to be converted into a machine-friendly object using the following function:

```{r prepare_for_enrichment}
go_terms <- prepare_for_enrichment(go$terms, go$mapping, exmpl_all,
                                   feature_name = "gene_symbol")
```

`exmpl_all` is an example of gene background - a vector with gene symbols related to all detections in an imaginary RNA-seq experiment. Since different datasets use different features (gene id, gene symbol, protein id), the column name containing features in `go$mapping` needs to be specified using `feature_name = "gene_symbol"`. The resulting object, `go_terms`, is a data structure containing all the mappings in a quickly accessible form. From this point on, `go_terms` can be employed to perform multiple functional enrichment analyses on various gene selections.

## Functional enrichment

The package includes two pre-defined gene sets. `exmpl_all` contains all background gene symbols, while `exmpl_sel` comprises the genes of interest. To perform functional enrichment analysis on the selected genes, you can use the following single, efficient function call:

```{r enrichment}
enr <- functional_enrichment(exmpl_all, exmpl_sel, go_terms)
```

## The output

The result of `functional_enrichment` is a tibble with enrichment results.

```{r enrichment_result}
enr |>
  head(10)
```

The columns are as follows

 - `N_with`: The number of features (genes) associated with this term in the background of all genes.
 - `n_with_sel`: The number of features associated with this term in the selection.
 - `n_expect`: The expected number of features associated with this term under the null hypothesis (terms are randomly distributed).
 - `enrichment`: The ratio of observed to expected.
 - `odds_ratio`: The effect size, represented by the odds ratio from the contingency table.
 - `ids`: The identifiers of features with the term in the selection.
 - `p_value`: The raw p-value from the hypergeometric distribution.
 - `p_adjust`: The p-value adjusted for multiple tests using the Benjamini-Hochberg approach.

# Interactive Example

A small Shiny app is included in the package to demonstrate the usage of *fenr* in an interactive environment. All time-consuming data loading and preparation tasks are performed before the app is launched.

```{r interactive_prepare, eval=FALSE}
data(yeast_de)
term_data <- fetch_terms_for_example(yeast_de)
```
 
`yeast_de` is the result of differential expression (using `edgeR`) on a subset of 6+6 replicates from [Gierlinski et al. (2015)](https://academic.oup.com/bioinformatics/article/31/22/3625/240923).

The function `fetch_terms_for_example` uses `fetch_*` functions from *fenr* to download and process data from *GO*, *Reactome* and *KEGG*. The step-by-step process of data acquisition can be examined by viewing the function by typing `fetch_terms_for_example` in the R console. The object `term_data` is a named list of `fenr_terms` objects, one for each ontology.

After completing the slow tasks, you can start the Shiny app by running:

```{r shiny_app, eval=FALSE}
enrichment_interactive(yeast_de, term_data)
```

To quickly see how *fenr* works an example can be loaded directly from GitHub:

```
shiny::runGitHub("bartongroup/fenr-shiny-example")
```

# Comparison to other functional enrichment packages

There's no shortage of R packages dedicated to functional enrichment analyses, including *topGO*, *GOstats*, *clusterProfiler*, and *GOfuncR*. Many of these packages provide comprehensive analyses tailored to a single gene selection. While they offer detailed insights, their extensive computational requirements often render them slower.

To illustrate, let's take the example of the *topGO* package.


```{r topGO}
suppressPackageStartupMessages({
  library(topGO)
})

# Preparing the gene set
all_genes <- setNames(rep(0, length(exmpl_all)), exmpl_all)
all_genes[exmpl_all %in% exmpl_sel] <- 1

# Define the gene selection function
gene_selection <- function(genes) {
  genes > 0
}

# Mapping genes to GO, we use the go2gene downloaded above and convert it to a
# list
go2gene <- go$mapping |> 
  dplyr::select(gene_symbol, term_id) |> 
  dplyr::distinct() |> 
  dplyr::group_by(term_id) |> 
  dplyr::summarise(ids = list(gene_symbol)) |> 
  tibble::deframe()

# Setting up the GO data for analysis
go_data <- new(
  "topGOdata",
  ontology = "BP",
  allGenes = all_genes,
  geneSel = gene_selection,
  annot = annFUN.GO2genes,
  GO2genes = go2gene
)

# Performing the enrichment test
fisher_test <- runTest(go_data, algorithm = "classic", statistic = "fisher")
```

Though the enrichment test (`runTest`) is fairly fast, completing in about 200 ms on an Intel Core i7 processor, the preparatory step (`new` call) demands a more substantial time commitment, around 6 seconds. Considering each additional gene selection requires a similar amount of time, conducting exploratory interactive analyses on multiple gene sets becomes notably cumbersome.

*fenr* separates preparatory steps and gene selection and is designed with interactivity and speed in mind, particularly when users wish to execute rapid, multiple selection tests.

```{r fenr_speed, eval = FALSE}
# Setting up data for enrichment
go <- fetch_go(species = "sgd")
go_terms <- prepare_for_enrichment(go$terms, go$mapping, exmpl_all,
                                   feature_name = "gene_symbol")

# Executing the enrichment test
enr <- functional_enrichment(exmpl_all, exmpl_sel, go_terms)
```

While data preparation with *fenr* might take a bit longer (around 10 s), the enrichment test is fast, finishing in just about 50 ms. Each new gene selection takes a similar amount of time (with an obvious caveat that a larger gene selection would take more time to calculate).


# Session info {.unnumbered}

```{r session_info, echo=FALSE}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrichment.R
\name{prepare_for_enrichment}
\alias{prepare_for_enrichment}
\title{Prepare Term Data for Enrichment Analysis}
\usage{
prepare_for_enrichment(
  terms,
  mapping,
  all_features = NULL,
  feature_name = "gene_id"
)
}
\arguments{
\item{terms}{A tibble with at least two columns: \code{term_id} and
\code{term_name}. This tibble contains information about functional term
names and descriptions.}

\item{mapping}{A tibble with at least two columns, containing the mapping
between functional terms and features. One column must be named
\code{term_id}, while the other column should have a name specified by the
\code{feature_name} argument. For example, if \code{mapping} contains
columns \code{term_id}, \code{accession_number}, and \code{gene_symbol},
setting \code{feature_name = "gene_symbol"} indicates that gene symbols
will be used for enrichment analysis.}

\item{all_features}{A vector with all feature IDs used as the background for
enrichment. If not specified, all features found in \code{mapping} will be
used, resulting in a larger object size.}

\item{feature_name}{The name of the column in the \code{mapping} tibble to be
used as the feature identifier. For example, if \code{mapping} contains
columns \code{term_id}, \code{accession_number}, and \code{gene_symbol},
setting \code{feature_name = "gene_symbol"} indicates that gene symbols
will be used for enrichment analysis.}
}
\value{
An object of class \code{fenr_terms} required by
  \code{functional_enrichment}.
}
\description{
Process term data downloaded with the \code{fetch_*} functions, preparing it
for fast enrichment analysis using \code{functional_enrichment}.
}
\details{
This function takes two tibbles containing functional term information
(\code{terms}) and feature mapping (\code{mapping}), and converts them into
an object required by \code{functional_enrichment} for efficient analysis.
Terms and mapping can be generated with the database access functions
included in this package, such as \code{fetch_reactome} or
\code{fetch_go_from_go}.
}
\examples{
\dontrun{
data(exmpl_all)
go <- fetch_go(species = "sgd")
go_terms <- prepare_for_enrichment(go$terms, go$mapping, exmpl_all,
                                   feature_name = "gene_symbol")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{exmpl_all}
\alias{exmpl_all}
\title{Example set of background genes.}
\format{
A character vector with 6985 elements.
}
\usage{
data(exmpl_all)
}
\value{
A vector with background gene names.
}
\description{
A set of gene names for proteins that could be detected in a typical
proteomics experiment on yeast samples.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/caching.R
\name{remove_cache}
\alias{remove_cache}
\title{Remove all cache}
\usage{
remove_cache(ask = TRUE)
}
\arguments{
\item{ask}{Logical, whether to ask user for confirmation.}
}
\value{
TRUE if successfully removed.
}
\description{
This function will remove all cached data used by `fenr`. The user will be
prompted for confirmation. Use only when you suspect the cache was corrupted.
Use with caution!
}
\examples{
\dontrun{
remove_cache()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{get_term_features}
\alias{get_term_features}
\title{Get features annotated with a given term.}
\usage{
get_term_features(term_data, term_id)
}
\arguments{
\item{term_data}{An object class \code{fenr_terms}, created by
\code{prepare_for_enrichment}.}

\item{term_id}{A string with a functional term ID.}
}
\value{
A character vector containing feature IDs annotated with the term ID.
}
\description{
Get features annotated with a given term.
}
\examples{
\dontrun{
go_data <- fetch_go(species = "sgd")
go_terms <- prepare_for_enrichment(go_data$terms, go_data$mapping, feature = "gene_symbol")
feats <- get_term_features(go_terms, "GO:0000001")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bioplanet.R
\name{fetch_bp}
\alias{fetch_bp}
\title{Get functional term data from BioPlanet}
\usage{
fetch_bp(use_cache = TRUE, on_error = c("stop", "warn", "ignore"))
}
\arguments{
\item{use_cache}{Logical, if TRUE, the remote file will be cached locally.}

\item{on_error}{A character string indicating the error handling strategy:
either "stop" to halt execution, "warn" to issue a warning and return
`NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".}
}
\value{
A list with \code{terms} and \code{mapping} tibbles.
}
\description{
Download term information (term ID and name) and gene-pathway mapping
(NCBI gene ID, gene symbol and pathway ID) from BioPlanet.
}
\examples{
bioplanet_data <- fetch_bp(on_error = "warn")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reactome.R
\name{fetch_reactome}
\alias{fetch_reactome}
\title{Get functional term data from Reactome}
\usage{
fetch_reactome(
  species,
  source = c("ensembl", "api", "gene_association"),
  use_cache = TRUE,
  on_error = c("stop", "warn", "ignore")
)
}
\arguments{
\item{species}{Reactome species designation, for example "Homo sapiens" for
human. Full list of available species can be found using
\code{fetch_reactome_species()}.}

\item{source}{How to download the mapping. If 'ensembl' or
'gene_association', one mapping file provided by Reactome will be
downloaded, if 'api', then Reactome API will be used. See details.}

\item{use_cache}{Logical, if TRUE, the remote file will be cached locally.}

\item{on_error}{A character string indicating the error handling strategy:
either "stop" to halt execution, "warn" to issue a warning and return
`NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".}
}
\value{
A list with \code{terms} and \code{mapping} tibbles
}
\description{
Download term information (pathway ID and name) and gene-pathway mapping
(Ensembl gene ID or gene symbol and pathway ID) from Reactome.
}
\details{
Reactome makes mapping between Ensembl ID and pathway ID available
  in form of one downloadable file. This mapping contains gene symbols as
  well. Also, a gene association file with mapping between UniProt accession
  number, gene symbol and Reactome term is available. If \code{source =
  "ensembl"} or \code{source =
  "gene_association"} is set, one large file will be downloaded and parsed.
  If \code{source = "api"} is set, then Reactome APIs will be interrogated
  for each pathway available. This method is considerably slower, especially
  for large genomes. However, gene association file contains far fewer
  mappings than can be extracted using API. If gene symbols are needed, we
  recommend using  \code{source = "ensembl"}.
}
\examples{
reactome_data <- fetch_reactome("Saccharomyces cerevisiae", on_error = "warn")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/go.R
\name{fetch_go_species}
\alias{fetch_go_species}
\title{Find all species available from geneontology.org}
\usage{
fetch_go_species(on_error = c("stop", "warn", "ignore"))
}
\arguments{
\item{on_error}{A character string indicating the error handling strategy:
either "stop" to halt execution, "warn" to issue a warning and return
`NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".
This argument is retained for API compatibility; the current implementation
reads a packaged data table rather than querying a remote server.}
}
\value{
A tibble with columns \code{species} and \code{designation}.
}
\description{
This function returns a fixed table of Gene Ontology species designations
included with the package. It is a temporary compatibility patch for legacy
Gene Ontology GAF file names such as \file{goa_human.gaf.gz},
\file{mgi.gaf.gz}, and \file{sgd.gaf.gz}. These legacy file names are still
served by Gene Ontology, but they are no longer listed on the annotation
downloads page that this function previously scraped.
}
\details{
The returned table may therefore become stale if Gene Ontology removes or
changes the legacy download paths. A future version of \pkg{fenr} may replace
this snapshot with a different discovery mechanism or with the newer
Gene Ontology annotation file names.
}
\examples{
go_species <- fetch_go_species(on_error = "warn")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrichment.R
\name{functional_enrichment}
\alias{functional_enrichment}
\title{Fast Functional Enrichment}
\usage{
functional_enrichment(feat_all, feat_sel, term_data, feat2name = NULL)
}
\arguments{
\item{feat_all}{A character vector with all feature identifiers, serving as
the background for enrichment.}

\item{feat_sel}{A character vector with feature identifiers in the selection.}

\item{term_data}{An object of class \code{fenr_terms}, created by
\code{prepare_for_enrichment}.}

\item{feat2name}{An optional named list to convert feature IDs into feature
names.}
}
\value{
A tibble with enrichment results, providing the following information
  for each term:
  \itemize{
    \item{\code{N_with} - number of features with this term among all features}
    \item{\code{n_with_sel} - number of features with this term in the selection}
    \item{\code{n_expect} - expected number of features with this term in the selection,
      under the null hypothesis that terms are mapped to features randomly}
    \item{\code{enrichment} - ratio of n_with_sel / n_expect}
    \item{\code{odds_ratio} - odds ratio for enrichment; is infinite when all
      features with the given term are in the selection}
    \item{\code{p_value} - p-value from a single hypergeometric test}
    \item{\code{p_adjust} - p-value adjusted for multiple tests using the
      Benjamini-Hochberg approach}
  }.
}
\description{
Perform fast functional enrichment analysis based on the hypergeometric
distribution. Designed for use in interactive applications.
}
\details{
This function carries out functional enrichment analysis on a
  selection of features (e.g., differentially expressed genes) using the
  hypergeometric probability distribution (Fisher's exact test). Features can
  be genes, proteins, etc. The \code{term_data} object contains functional
  term information and feature-term mapping.
}
\examples{
\dontrun{
data(exmpl_all, exmpl_sel)
go <- fetch_go(species = "sgd")
go_terms <- prepare_for_enrichment(go$terms, go$mapping, exmpl_all, feature_name = "gene_symbol")
enr <- functional_enrichment(exmpl_all, exmpl_sel, go_terms)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{yeast_de}
\alias{yeast_de}
\title{Differential expression results for yeast RNA-seq.}
\format{
A tibble with 5 columns
}
\usage{
data(yeast_de)
}
\value{
Results for differential expression for yeast RNA-seq.
}
\description{
A subset of 6 + 6 replicates was selected from data set reported in
https://doi.org/10.1093/bioinformatics/btv425
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wiki.R
\name{fetch_wiki}
\alias{fetch_wiki}
\title{Get functional term data from WikiPathways}
\usage{
fetch_wiki(
  species,
  databases = c("Ensembl", "Entrez Gene", "HGNC", "HGNC Accession number",
    "Uniprot-TrEMBL"),
  types = c("GeneProduct", "Protein", "Rna", "RNA"),
  on_error = c("stop", "warn", "ignore")
)
}
\arguments{
\item{species}{WikiPathways species designation, for example "Homo sapiens"
for human. Full list of available species can be found using
\code{fetch_wiki_species()}.}

\item{databases}{A character vector with database names to pre-filter mapping
data. See details. Full result will be returned if NULL.}

\item{types}{A character vector with types of entities to pre-filter mapping
data. See details. Full result will be returned if NULL.}

\item{on_error}{A character string indicating the error handling strategy:
either "stop" to halt execution, "warn" to issue a warning and return
`NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".}
}
\value{
A list with \code{terms} and \code{mapping} tibbles.
}
\description{
Download term information (pathway ID and name) and gene-pathway mapping
(gene symbol and pathway ID) from WikiPathways.
}
\details{
WikiPathways contain mapping between pathways and a variety of
  entities from various databases. Typically a gene symbol is returned in
  column \code{text_label} and some sort of ID in column \code{id}, but this
  depends on the species and databases used. For gene/protein enrichment,
  these should be filtered to contain gene symbols only. This can be done by
  selecting a desired databases and types. The default values for parameters
  \code{databases} and \code{types} attempt to select information from generic
  databases, but there are organism-specific databases not included in the
  selection. We suggest to run this function with \code{databases = NULL,
  types = NULL} to see what types and databases are available before making
  selection.
}
\examples{
wiki_data <- fetch_wiki("Bacillus subtilis", on_error = "warn")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wiki.R
\name{fetch_wiki_species}
\alias{fetch_wiki_species}
\title{List of available WikiPathways species}
\usage{
fetch_wiki_species(on_error = c("stop", "warn", "ignore"))
}
\arguments{
\item{on_error}{A character string indicating the error handling strategy:
either "stop" to halt execution, "warn" to issue a warning and return
`NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".}
}
\value{
A character vector with species names used by WikiPathways.
}
\description{
List of available WikiPathways species
}
\examples{
spec <- fetch_wiki_species(on_error = "warn")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reactome.R
\name{fetch_reactome_species}
\alias{fetch_reactome_species}
\title{List of available Reactome species}
\usage{
fetch_reactome_species(on_error = c("stop", "warn", "ignore"))
}
\arguments{
\item{on_error}{A character string indicating the error handling strategy:
either "stop" to halt execution, "warn" to issue a warning and return
`NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".}
}
\value{
A tibble with species names used by Reactome.
}
\description{
List of available Reactome species
}
\examples{
re <- fetch_reactome_species(on_error = "warn")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kegg.R
\name{fetch_kegg_species}
\alias{fetch_kegg_species}
\title{Find all species available from KEGG}
\usage{
fetch_kegg_species(on_error = c("stop", "warn", "ignore"))
}
\arguments{
\item{on_error}{A character string indicating the error handling strategy:
either "stop" to halt execution, "warn" to issue a warning and return
`NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".}
}
\value{
A tibble, in which column \code{designation} contains species
  designations used in function \code{fetch_kegg}.
}
\description{
Find all species available from KEGG
}
\examples{
spe <- fetch_kegg_species(on_error = "warn")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{get_feature_terms}
\alias{get_feature_terms}
\title{Get terms annotating a given feature.}
\usage{
get_feature_terms(term_data, feature_id)
}
\arguments{
\item{term_data}{An object class \code{fenr_terms}, created by
\code{prepare_for_enrichment}.}

\item{feature_id}{A string with a feature ID}
}
\value{
A character vector containing functional term IDs annotating given feature.
}
\description{
Get terms annotating a given feature.
}
\examples{
\dontrun{
go_data <- fetch_go(species = "sgd")
go_terms <- prepare_for_enrichment(go_data$terms, go_data$mapping, feature = "gene_symbol")
trms <- get_feature_terms(go_terms, "GEM1")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{go}
\alias{go}
\title{GO-terms data downloaded for the vignette.}
\format{
A list of two tibbles
}
\usage{
data(go)
}
\value{
Contains GO-term descriptions and gene mapping.
}
\description{
Downloaded using \code{go <- fetch_go(species = "sgd")}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/go.R
\name{fetch_go}
\alias{fetch_go}
\title{Get Gene Ontology (GO) data}
\usage{
fetch_go(
  species = NULL,
  dataset = NULL,
  use_cache = TRUE,
  on_error = c("stop", "warn", "ignore")
)
}
\arguments{
\item{species}{(Optional) Species designation. Examples are \code{goa_human}
for human, \code{mgi} for mouse, or \code{sgd} for yeast. Full list of
available species can be obtained using \code{fetch_go_species} - column
\code{designation}. This argument is used when fetching data from the Gene
Ontology database.}

\item{dataset}{(Optional) A string representing the dataset passed to
Ensebml's Biomart, e.g. 'scerevisiae_gene_ensembl'. To see the different
datasets available within a biomaRt you can e.g. do: mart <-
biomaRt::useEnsembl(biomart = "ensembl"), followed by
biomaRt::listDatasets(mart).}

\item{use_cache}{Logical, if TRUE, the remote data will be cached locally.}

\item{on_error}{A character string indicating the error handling strategy:
either "stop" to halt execution, "warn" to issue a warning and return
`NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".}
}
\value{
A list with \code{terms} and \code{mapping} tibbles.
}
\description{
This function downloads term information (GO term ID and name) and gene-term
mapping (gene ID, symbol, and GO term ID) from either the Ensembl database
(using BioMart) or the Gene Ontology database (using GAF files), depending on
the provided argument.
}
\details{
If \code{species} is provided, mapping from a Gene Ontology GAF file
  will be downloaded. GAF files contain more generic information than gene
  symbols. In this function, the third column of the GAF file (DB Object
  Symbol) is returned as \code{gene_symbol}, but, depending on the
  \code{species} argument it can contain other entities, e.g. RNA or protein
  complex names. Similarly, the eleventh column of the GAF file (DB Object
  Synonym) is returned as \code{gene_id}. It is up to the user to select
  the appropriate database.

  Alternatively, if \code{dataset} is provided, mapping will be downloaded
  from Ensembl database. It will gene symbol and Ensembl gene ID.
}
\examples{
# Fetch GO data from Ensembl
go_data_ensembl <- fetch_go(dataset = "scerevisiae_gene_ensembl", on_error = "warn")
# Fetch GO data from Gene Ontology
go_data_go <- fetch_go(species = "sgd", on_error = "warn")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{go_species}
\alias{go_species}
\title{GO species}
\format{
A tibble
}
\usage{
data(go_species)
}
\value{
Contains species available through Gene Ontology
}
\description{
Downloaded using \code{go_species <- fetch_go_species()}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interactive_example.R
\name{enrichment_interactive}
\alias{enrichment_interactive}
\title{Small Shiny app serving as example for fast enrichment}
\usage{
enrichment_interactive(de, term_data)
}
\arguments{
\item{de}{Differential expression results, \code{yeast_de} attached to this
package can be used.}

\item{term_data}{A list of \code{fenr_terms} objects containing functional
data for various ontologies. \code{fetch_terms_for_example} can be used to
create this object.}
}
\value{
An interactive Shiny app
}
\description{
Small Shiny app serving as example for fast enrichment
}
\examples{
\dontrun{
data(yeast_de)
term_data <- fetch_terms_for_example(yeast_de)
  enrichment_interactive(yeast_de, term_data)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fenr-package.R
\docType{package}
\name{fenr-package}
\alias{fenr}
\alias{fenr-package}
\title{fenr: Fast functional enrichment for interactive applications}
\value{
Package fenr.
}
\description{
This R package provides a fast and efficient method for functional enrichment
analysis, optimized for speed and designed for use in interactive
applications, such as *Shiny* apps.

To learn more about fenr, start with the vignette: vignette("fenr").
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/bartongroup/fenr}
  \item Report bugs at \url{https://github.com/bartongroup/fenr/issues}
}

}
\author{
\strong{Maintainer}: Marek Gierlinski \email{M.Gierlinski@dundee.ac.uk} (\href{https://orcid.org/0000-0001-9149-3514}{ORCID})

Authors:
\itemize{
  \item Marek Gierlinski \email{M.Gierlinski@dundee.ac.uk} (\href{https://orcid.org/0000-0001-9149-3514}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{exmpl_sel}
\alias{exmpl_sel}
\title{Example set of selected genes.}
\format{
A character vector with 21 elements.
}
\usage{
data(exmpl_sel)
}
\value{
A vector with selected gene names.
}
\description{
A set of gene names manually selected to illustrate functional enrichment.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kegg.R
\name{fetch_kegg}
\alias{fetch_kegg}
\title{Get functional term data from KEGG}
\usage{
fetch_kegg(species, batch_size = 10, on_error = c("stop", "warn", "ignore"))
}
\arguments{
\item{species}{KEGG species code, for example "hsa" for human. The
full list of available KEGG species can be found by using \code{fetch_kegg_species}.}

\item{batch_size}{Number of pathways sent to KEGG database in one query. The
maximum allowed is 10.}

\item{on_error}{A character string indicating the error handling strategy:
either "stop" to halt execution, "warn" to issue a warning and return
`NULL` or "ignore" to return `NULL` without warnings. Defaults to "stop".}
}
\value{
A list with \code{terms} and \code{mapping} tibbles.
}
\description{
Download information (pathway ID and name) and gene-pathway mapping (entrez
gene ID, gene symbol and pathway ID) from KEGG. Gene symbols are extracted
from gene descriptions. For some species (e.g. yeast), gene symbols are
returned instead of entrez IDs and not in gene description.
}
\examples{
kegg_data <- fetch_kegg("mge", on_error = "warn")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interactive_example.R
\name{fetch_terms_for_example}
\alias{fetch_terms_for_example}
\title{Create term data for interactive example}
\usage{
fetch_terms_for_example(de)
}
\arguments{
\item{de}{Differential expression results, use \code{yeast_de} data attached
to this package.}
}
\value{
A list of objects containing functional terms for GO and Reactome.
}
\description{
Create term data for interactive example
}
\examples{
\dontrun{
data(yeast_de)
term_data <- fetch_terms_for_example(yeast_de)
}
}
